#/usr/bin/env python

#    This file is part of Diamond.
#
#    Diamond is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Diamond is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Diamond.  If not, see <http://www.gnu.org/licenses/>.

import base64
import bz2
import copy
import cPickle as pickle
import cStringIO as StringIO
import re
import zlib
from lxml import etree
import sys

import gobject

import debug
import choice
import mixedtree

class Tree(gobject.GObject):
  """This class maps pretty much 1-to-1 with an xml tree.
     It is used to represent the options in-core."""

  __gsignals__ = { "on-set-data" : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, (str,)),
                   "on-set-attr"  : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, (str, str))}
  
  def __init__(self, name="", schemaname="", attrs={}, children=None, cardinality='', datatype=None, doc=None):
    gobject.GObject.__init__(self)

    # name: the element name in the options XML
    # e.g. "fluidity_options"
    self.name = name
    
    # schemaname: the label given to it in the Xvif parsing of the schema
    # this is necessary to walk the tree to see what possible valid
    # children this node could have
    # e.g. "0:elt" for the root node.
    self.schemaname = schemaname

    # Any children?
    if children is None:
      self.children = []
    else:
      self.children = children

    # The cardinality of a node is
    # how many you must/can have, e.g.
    # "exactly one", "zero or one", "any amount", etc.
    # This is set by Schema.valid_children for candidate
    # nodes in the tree, you see.
    # Possible choices: '' '?' '*' '+'
    # with the usual regex meanings.
    self.cardinality = cardinality

    # Used for Optional or ZeroOrMore
    # trees. False means it is present but inactive.
    # must be set if cardinality is changed!
    self.set_default_active()

    # Any documentation associated with this node?
    self.doc = doc

    # What is the parent of this tree?
    # None means the root node.
    self.parent = None

    # Does this node require attention from the user?
    self.valid = False

    # The datatype that this tree stores and the data stored
    if isinstance(datatype, tuple) and len(datatype) == 1:
      self.datatype = "fixed"
      self.data = datatype[0]
    else:
      self.datatype = datatype
      self.data = None

    # The attributes of the tree
    self.attrs = {}
    for key in attrs.keys():
      if isinstance(attrs[key][0], tuple) and len(attrs[key][0]) == 1:
        self.attrs[key] = ("fixed", attrs[key][0][0])
      else:
        self.attrs[key] = attrs[key]

    self.recompute_validity()

  def set_attr(self, attr, val):
    """Set an attribute."""
    (datatype, curval) = self.attrs[attr]
    (invalid, newdata) = self.valid_data(datatype, val)
    if invalid:
      raise Exception, "invalid data: (%s, %s)" % (datatype, val)
    self.attrs[attr] = (datatype, newdata)
    self.recompute_validity()
    self.emit("on-set-attr", attr, val)

  def get_attr(self, attr):
    """Get an attribute."""
    (datatype, curval) = self.attrs[attr]
    return curval

  def get_attrs(self):
    """Get all attributes"""
    return self.attrs

  def set_data(self, data):
    (invalid, data) = self.valid_data(self.datatype, data)
    if invalid:
      raise Exception, "invalid data: (%s, %s)" % (str(self.datatype), data)
    self.data = data
    self.recompute_validity()
    self.emit("on-set-data", data)

  def valid_data(self, datatype, data):
    if datatype is None:
      raise Exception, "datatype is None!"

    elif datatype == "fixed":
      raise Exception, "datatype is fixed!"

    datatypes_to_check = []

    if isinstance(datatype, tuple):
      if isinstance(datatype[0], tuple):
        fixed_values = datatype[0]
      else:
        fixed_values = datatype
      if data in fixed_values:
        return (False, data)
      else:
        if not isinstance(datatype[0], tuple):
          return (True, data)
        datatypes_to_check = list(datatype[1:])
    else:
      datatypes_to_check = [datatype]

    for datatype in datatypes_to_check:
      try:
        tempval = datatype(data)
        if isinstance(tempval, str):
          data = tempval
        return (False, data)
      except:
        pass

    return (True, data)

  def validity_check(self, datatype, data):
    """
    Check to see if the supplied data with supplied type can be stored in a
    tree.Tree.
    """

    (invalid, new_data) = self.valid_data(datatype, data)
    if not invalid and isinstance(new_data, str) and new_data != "":
      if new_data != data and self.validity_check(datatype, new_data) is None:
        return None
      else:
        return new_data
    else:
      return None

  def copy(self):
    new_copy = Tree()
    for attr in ["attrs", "name", "schemaname", "doc", "cardinality", "datatype", "data", "active", "valid"]:
      setattr(new_copy, attr, copy.copy(getattr(self, attr)))

    new_copy.parent = self.parent
    new_copy.children = []

    return new_copy

  def recompute_validity(self):

    new_valid = True

    # if any children are invalid,
    # we are invalid too
    for child in self.children:
      if child.active is False: continue

      if child.__class__ is choice.Choice:
        child = child.get_current_tree()

      if child.valid is False:
        new_valid = False

    # if any attributes are unset,
    # we are invalid.
    for attr in self.attrs.keys():
      (datatype, val) = self.attrs[attr]
      if not datatype is None and val is None:
        new_valid = False

    # if we're supposed to have data and don't,
    # we are invalid.
    if self.datatype is not None:
      if not hasattr(self, "data"):
        new_valid = False

      if self.data is None:
        new_valid = False

    # so we are valid.
    # in either case, let's let the parent know.
    self.valid = new_valid

    if self.parent is not None:
      self.parent.recompute_validity()

  def find_or_add(self, treelist):
    """Append a child node to this node in the tree.
       If it already exists, make tree point to it."""

    outlist = []

    for tree in treelist:
      found = False
      for t in self.children:
        if t.schemaname == tree.schemaname:
          outlist.append(t)
          found = True

      if not found:
        tree.set_parent(self)

        self.children.append(tree)
        tree.recompute_validity()
        outlist.append(tree)

      for tree in outlist:
        if tree.cardinality in ['+', '*']:
          inactive_list = [x for x in outlist if x.schemaname == tree.schemaname and x.active is False]
          if len(inactive_list) > 0: continue
          else:
            new_tree = self.add_inactive_instance(tree)
            outlist.insert(outlist.index(tree)+1, new_tree)

    return outlist

  def write(self, filename):
    if isinstance(filename, str):
      file = open(filename, "w")
    else:
      file = filename

    xmlTree=etree.tostring(self.write_core(None), pretty_print = True, xml_declaration = True, encoding="utf-8")
    
    file.write(xmlTree)

  def write_core(self, parent):
    """Write to XML; this is the part that recurses"""

    sub_tree=etree.Element(self.name)
    
    for key in self.attrs:
      val = self.attrs[key]
      output_val = val[1]
      if output_val is not None:
        sub_tree.set(unicode(key), unicode(output_val))
  
    for child in self.children:
      if child.active is True:
        child.write_core(sub_tree)
        
    if self.data is not None:
      sub_tree.text = unicode(self.data)
      
    if parent is not None:
      parent.append(sub_tree)

    return sub_tree

  def pickle(self):
    if hasattr(self, "xmlnode"):
      del self.xmlnode

    return base64.b64encode(bz2.compress(pickle.dumps(self)))

  def unpickle(self, pick):
    return pickle.loads(bz2.decompress(base64.b64decode(pick)))

  def print_str(self):
    s = "name: %s at %s\n" % (self.name, hex(id(self)))
    s = s + "schemaname: %s\n" % self.schemaname
    s = s + "attrs: %s\n" % self.attrs
    s = s + "children: %s\n" % self.children
    if self.parent is not None:
      s = s + "parent: %s %s at %s\n" % (self.parent.__class__, self.parent.name, hex(id(self.parent)))
    else:
      s = s + "parent: %s at %s\n" % (self.parent.__class__, hex(id(self.parent)))
    s = s + "datatype: %s\n" % str(self.datatype)
    s = s + "data: %s\n" % str(self.data)
    s = s + "cardinality: %s\n" % self.cardinality
    s = s + "active: %s\n" % self.active
    s = s + "valid: %s\n" % self.valid
    return s

  def set_default_active(self):
    self.active = True
    if self.cardinality == '?' or self.cardinality == '*':
      self.active = False

  def count_children_by_schemaname(self, schemaname):
    count = len(filter(lambda x: x.schemaname == schemaname, self.children))
    return count

  def get_children_by_schemaname(self, schemaname):
    return filter(lambda x: x.schemaname == schemaname, self.children)

  def delete_child_by_ref(self, ref):
    self.children.remove(ref)

  def add_inactive_instance(self, tree):
    for t in self.children:
      if t.schemaname == tree.schemaname and t.active is False:
        return t

    new_tree = tree.copy()
    new_tree.active = False
    if new_tree.__class__ is Tree:
      new_tree.children = []
    new_tree.parent = tree.parent
    self.children.insert(self.children.index(tree)+1, new_tree)
    return new_tree

  def print_recursively(self, indent=""):
    s = self.__str__()
    debug.dprint(indent + ' ' + s.replace('\n', '\n' + indent + ' '), 0, newline = False)
    debug.dprint("", 0)
    for i in range(len(self.children)):
      if isinstance(self.children[i], Tree):
        self.children[i].print_recursively(indent + ">>")
      elif isinstance(self.children[i], choice.Choice):
        ref = self.children[i].get_current_tree()
        ref.print_recursively(indent + ">>")
      if i < len(self.children) - 1:
        debug.dprint("", 0)

    return

  def add_children(self, schema):
    l = schema.valid_children(self)
    l = self.find_or_add(l)
    for child in self.children:
      child.add_children(schema)

  def matches(self, text, case_sensitive = False):
    if case_sensitive:
      text_re = re.compile(text)
    else:
      text_re = re.compile(text, re.IGNORECASE)

    if not text_re.search(self.name) is None:
      return True

    if not self.doc is None:
      if not text_re.search(self.doc) is None:
        return True

    for key in self.attrs:
      if not text_re.search(key) is None:
        return True
      if not self.get_attr(key) is None:
        if not text_re.search(self.get_attr(key)) is None:
          return True

    if not self.data is None:
      if not text_re.search(self.data) is None:
        return True

    return False

  def get_current_tree(self):
    return self

  def get_possible_names(self):
    return [self.name]

  def set_parent(self, parent):
    self.parent = parent

  def find_tree(self, name):
    if name == self.name:
      return self
    else:
      raise Exception, "ban the bomb"

  def choices(self):
    return [self]

  def is_comment(self):
    """
    Test whether the given node is a comment node.
    """

    if not self.name == "comment":
      return False

    if not self.attrs == {}:
      return False

    if not self.children == []:
      return False

    if not self.datatype is str:
      return False

    if not self.cardinality == "?":
      return False

    return True

  def get_comment(self):
    """
    Return the first comment found as a child of the supplied node, or None if
    none found.
    """

    for child in self.children:
      if child.is_comment():
        return child

    return None

  def is_tensor(self, geometry_dim_tree):
    return False

  def is_code(self):
    """
    Perform a series of tests on the current Tree, to determine if
    it is intended to be used to store code data.
    """

    try:
      lang = self.get_attr("language")
      if lang == "python":
        return True
    except:
      pass

    return False

  def get_code_language(self):
    """
    Assuming this is a code snippet return the language.
    """

    if not self.is_code():
      return "python"

    try:
      lang = self.get_attr("language")
      return lang
    except:   
      return "python"

  def get_display_name(self):
    """
    This is a fluidity hack, allowing the name displayed in the treeview on the
    left to be different to the element name. If it has an attribute name="xxx",
    element_tag (xxx) is displayed.
    """
    
    name = self.get_name()

    if name is None:
      return self.name
    else:
      return self.name + " (" + name + ")"

  def get_name(self):
    if "name" in self.attrs:
      name = self.attrs["name"][1]
      return name

    return None

  def get_children(self):
    return self.children
  
  def get_choices(self):
    return [self]

  def is_hidden(self):
    """
    Tests whether the supplied tree should be hidden in view.
    """
    return self.is_comment() or self.name in ["integer_value", "real_value", "string_value", "logical_value"]

  def get_name_path(self, leaf = True):
    name = self.get_display_name() if leaf else self.get_name()

    if self.parent is None:
      return name                            
    else:

      pname = self.parent.get_name_path(False)

      if name is None:
        return pname
      elif pname is None:
        return name
      else:
        return pname + "/" + name

  def get_mixed_data(self):
    integers = [child for child in self.children if child.name == "integer_value"]
    reals    = [child for child in self.children if child.name == "real_value"]
    logicals = [child for child in self.children if child.name == "logical_value"]
    strings  = [child for child in self.children if child.name == "string_value"]

    child = None
    if len(integers) > 0:
      child = integers[0]
    if len(reals) > 0:
      child = reals[0]
    if len(logicals) > 0:
      child = logicals[0]
    if len(strings) > 0:
      child = strings[0]

    if child is None:
      return self
    else:
      return mixedtree.MixedTree(self, child)

  def is_sliceable(self):
    mixed = self.get_mixed_data()
    if isinstance(mixed, mixedtree.MixedTree):
      return True
    
    return (self.datatype is not None and self.datatype != "fixed") or self.attrs

  def __str__(self):
    return self.get_display_name()
  
  def __repr__(self):
    return self.get_name_path()

  def all_attrs_fixed(self):
    for attr in self.attrs:
      if self.attrs[attr][0] != "fixed":
        return False

    return True
 
gobject.type_register(Tree)




# Author: Daryl Harrison

# Enthought library imports.
from enthought.traits.api import Instance, String, List, Int
from enthought.traits.ui.api import View, Item
from enthought.tvtk.api import tvtk

# Local imports.
from enthought.mayavi.core.file_data_source import FileDataSource
from enthought.mayavi.core.pipeline_info import PipelineInfo
from enthought.mayavi.core.traits import DEnum
from os import path
from numpy import array, append, fromstring
from re import compile, MULTILINE

######################################################################
# `TriangleReader` class
######################################################################
class TriangleReader(FileDataSource):
    """
    Reader for the Triangle file formats:
        2D <http://www.cs.cmu.edu/~quake/triangle.html>
            Supports opening .egde files to construct a surface mesh comprised of lines
            and .ele files to construct a solid mesh comprised of triangles.
        3D <http://tetgen.berlios.de/fformats.html>
            Supports opening .face files to construct a surface mesh comprised of triangles
            and .ele files to construct a solid mesh comprised of tetrahedra.
    Outputs an unstructured grid dataset.
    """

    # The version of this class.  Used for persistence.
    __version__ = 0

    # Information about what this object can produce.
    output_info = PipelineInfo(datasets=['unstructured_grid'],
                               attribute_types=['any'],
                               attributes=['scalars'])

    # The active point scalar name.
    point_scalars_name = DEnum(values_name='_point_scalars_list',
                               desc='scalar point data attribute to use')

    # The active cell scalar name.
    cell_scalars_name = DEnum(values_name='_cell_scalars_list',
                               desc='scalar cell data attribute to use')

    ########################################
    # Private traits.

    # These private traits store the list of available data
    # attributes.  The non-private traits use these lists internally.
    _cell_scalars_list = List(String)
    _point_scalars_list = List(String)

    # The VTK dataset to manage.
    _grid = Instance(tvtk.UnstructuredGrid, args=(), allow_none=False)

    # The basename of the file which has been loaded.
    _basename = String

    # Indicates whether nodes are numbered from 0 or 1 (the file
    # format allows both).
    _numbered_from = Int

    # This filter allows us to change the attributes of the data
    # object and will ensure that the pipeline is properly taken care
    # of.
    _assign_attribute = Instance(tvtk.AssignAttribute, args=(), allow_none=False)

    ########################################
    # The view.

    view = View(Item(name='point_scalars_name'),
                Item(name='cell_scalars_name'))

    ########################################
    # `FileDataSource` interface.

    def initialize(self, base_file_name):
        split = path.splitext(base_file_name)
        self._basename = split[0]
        extension = split[1]

        self._assign_attribute.input = self._grid

        self._read_node_file()
        if (extension == '.face' or extension == '.edge'):
            self._read_face_edge_file(extension)
        else:
            self._read_ele_file()

        self.outputs = [self._assign_attribute.output]
        self.name = 'Triangle file (%s%s)' %(path.basename(self._basename), extension)

    ########################################
    # File reading methods.

    def _read_node_file(self):
        """
        Loads data from {basename}.node, and inserts points and point
        scalars into the unstructured grid.
        """
        file_name = '%s.node' %self._basename

        # Load all data.
        all_data = self._get_data(file_name)
        # Grab values from the first line of data file.
        points, dimensions, attributes, boundary_marker = map(int, all_data[0:4])
        # Reshape remainder of array.
        data_array = all_data[4:].reshape(points, 1+dimensions+attributes+boundary_marker)

        self._numbered_from = int(data_array[0][0])

        points_array = array(data_array[:, 1:(1+dimensions)], 'double')
        if (dimensions == 2):
            # Add a 0 to each point if it is 2D.
            points_array = array(map (lambda a: append(a, 0), points_array))
        self._grid.points = points_array

        for i in range(attributes):
            attribute_array = data_array[:, (i+dimensions+1):(i+dimensions+2)]
            self._add_attribute_array(attribute_array, i, 'point')

        if (boundary_marker):
            boundary_marker_array = data_array[:, (dimensions+attributes+1):(dimensions+attributes+2)]
            self._add_boundary_marker_array(boundary_marker_array, 'point')


    def _read_face_edge_file(self, extension):
        """
        Loads data from 2D {basename}.edge or 3D {basename}.face, and inserts
        triangle/line cells and cell scalars into the unstructured grid.
        """
        file_name = '%s%s' %(self._basename, extension)

        if (extension == '.edge'):
            # 2D. Expect two endpoints which form a line.
            npoints = 2
            cell_type = tvtk.Line().cell_type
        else:
            # 3D. Expect three points which form a triangle.
            npoints = 3
            cell_type = tvtk.Triangle().cell_type

        # Load all data.
        all_data = self._get_data(file_name)
        # Grab values from the first line of data file.
        faces_edges, boundary_marker = map(int, all_data[0:2])
        # Reshape remainder of array.
        data_array = all_data[2:].reshape(faces_edges, npoints+1+boundary_marker)

        nodes_array = data_array[:, 1:npoints+1] - self._numbered_from
        self._grid.set_cells(cell_type, nodes_array)

        if (boundary_marker):
            boundary_marker_array = data_array[:, npoints+1:npoints+2]
            self._add_boundary_marker_array(boundary_marker_array, 'cell')


    def _read_ele_file(self):
        """
        Loads data from {basename}.ele, and inserts triangle/tetrahedron cells
        and cell scalars into the unstructured grid.
        """
        file_name = '%s.ele' %self._basename

        # Load all data.
        all_data = self._get_data(file_name)
        # Grab values from the first line of data file.
        tet_tri, nodes_per_tet_tri, attributes =  map(int, all_data[0:3])
        # Reshape remainder of array.
        data_array = all_data[3:].reshape(tet_tri, 1+nodes_per_tet_tri+attributes)

        nodes_array = data_array[:, 1:(nodes_per_tet_tri+1)] - self._numbered_from

        if (nodes_per_tet_tri == 3):
            cell_type = tvtk.Triangle().cell_type
        else:
            cell_type = tvtk.Tetra().cell_type

        self._grid.set_cells(cell_type, nodes_array)

        for i in range(attributes):
            attribute_array = data_array[:, (i+nodes_per_tet_tri+1):(i+nodes_per_tet_tri+2)]
            self._add_attribute_array(attribute_array, i, 'cell')


    def _get_data(self, file_name):
        """
        Returns a 1D array containing all the data from the given file.
        """
        file = open(file_name)
        file_string = file.read()

        # Strip comments.
        pattern = compile('#.*?$', MULTILINE)
        file_string = pattern.sub('', file_string)

        # Load all data into array.
        return fromstring(file_string, dtype=float, sep=" ")

    ########################################
    # Unstructured grid construction
    # methods.

    def _add_attribute_array(self, attribute_array, i, type):
        """
        Adds the given attribute array to either point_data or
        cell_data of the unstructured grid.
        """
        attribute_array_name = 'Attribute %i' %i
        if (type == 'cell'): # .ele file attributes are of type Int
            tvtk_attribute_array = tvtk.IntArray(name=attribute_array_name)
            attribute_array = map(int, attribute_array)
        else:                # .node file attributes are of type Float
            tvtk_attribute_array = tvtk.FloatArray(name=attribute_array_name)
        tvtk_attribute_array.from_array(attribute_array)
        getattr(self._grid, '%s_data' %type).add_array(tvtk_attribute_array)
        getattr(self, '_%s_scalars_list' %type).append(attribute_array_name)

        if (i == 0):
            self._set_data_name(type, 'Attribute 0')


    def _add_boundary_marker_array(self, boundary_marker_array, type):
        """
        Adds the given boundary marker array to either point_data or
        cell_data of the unstructured grid.
        """
        boundary_marker_array_name = 'Boundary Marker'
        tvtk_boundary_marker_array = tvtk.IntArray(name=boundary_marker_array_name)
        tvtk_boundary_marker_array.from_array(boundary_marker_array)
        getattr(self._grid, '%s_data' %type).add_array(tvtk_boundary_marker_array)
        getattr(self, '_%s_scalars_list' %type).append(boundary_marker_array_name)
        self._set_data_name(type, 'Boundary Marker')

    ########################################
    # Methods taken and modified from
    # SetActiveAttribute filter.

    def _point_scalars_name_changed(self, value):
        self._set_data_name('point', value)


    def _cell_scalars_name_changed(self, value):
        self._set_data_name('cell', value)


    def _set_data_name(self, attr_type, value):
        """
        Sets the selected point or cell scalar to be active, and
        deactivates the scalar of the other type.
        """
        if value is None:
            return

        if (attr_type == 'point'):
            data = self._grid.point_data
            other_data = self._grid.cell_data
        else:
            data = self._grid.cell_data
            other_data = self._grid.point_data

        method = getattr(data, 'set_active_scalars')
        method(value)

        # Deactivate other attribute.
        method = getattr(other_data, 'set_active_scalars')
        method(None)

        self._assign_attribute.assign(value, 'SCALARS', attr_type.upper()+'_DATA')
        self._assign_attribute.update()

        # Fire an event, so the changes propagate.
        self.data_changed = True



#!/usr/bin/env python

#    This file is part of Diamond.
#
#    Diamond is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Diamond is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Diamond.  If not, see <http://www.gnu.org/licenses/>.

import plist

class MixedTree:
  def __init__(self, parent, child):
    """
    The .doc and .attrs comes from parent, and the .data comes from child. This
    is used to hide integer_value etc. from the left hand side, but for its data
    entry to show up on the right.
    """

    self.parent = parent
    self.child = child

    self.name = parent.name
    self.schemaname = parent.schemaname

    excluded_attrs = ["shape"]
    self.attrs = dict(self.parent.attrs.items() + [x for x in self.child.attrs.items() if x[0] not in excluded_attrs])

    self.children = parent.children
    self.datatype = child.datatype
    self.data = child.data
    self.doc = parent.doc
    self.active = parent.active

    return

  def set_attr(self, attr, val):
    if attr in self.parent.attrs:
      self.parent.set_attr(attr, val)
    elif attr in self.child.attrs:
      self.child.set_attr(attr, val)
    else:
      raise Exception, "Attribute not present in either parent or child!"

    return

  def get_attr(self, attr):
    if attr in self.parent.attrs:
      self.parent.get_attr(attr)
    elif attr in self.child.attrs:
      self.child.get_attr(attr)
    else:
      raise Exception, "Attribute not present in either parent or child!"

  def set_data(self, data):
    self.child.set_data(data)
    self.datatype = self.child.datatype
    self.data = self.child.data

    return

  def valid_data(self, datatype, data):
    return self.child.valid_data(datatype, data)

  def validity_check(self, datatype, data):
    return self.child.validity_check(datatype, data)

  def matches(self, text, case_sensitive = False):
    old_parent_data = self.parent.data
    self.parent.data = None
    parent_matches = self.parent.matches(text, case_sensitive)
    self.parent.data = old_parent_data

    if parent_matches:
      return True

    if case_sensitive:
      text_re = re.compile(text)
    else:
      text_re = re.compile(text, re.IGNORECASE)

    if not self.child.data is None and not text_re.search(self.child.data) is None:
      return True
    else:
      return False

  def is_comment(self):
    return self.parent.is_comment()

  def get_comment(self):
    return self.parent.get_comment()

  def is_tensor(self, geometry_dim_tree):
    """
    Perform a series of tests on the current MixedTree, to determine if
    it is intended to be used to store tensor or vector data.
    """

    # Check that a geometry is defined
    if geometry_dim_tree is None:
      return False

    # Check that this element has calculable and positive dimensions
      try:
        dim1, dim2 = self.tensor_shape(geometry_dim_tree)
        assert dim1 > 0
        assert dim2 > 0
      except AssertionError:
        return False

    # The element must have dim1, rank and shape attributes
    if "dim1" not in self.child.attrs.keys() \
       or "rank" not in self.child.attrs.keys() \
       or "shape" not in self.child.attrs.keys():
      return False

    # The dim1 and rank attributes must be of fixed type
    if self.child.attrs["dim1"][0] != "fixed" or self.child.attrs["rank"][0] != "fixed":
      return False

    if "dim2" in self.child.attrs.keys():
      # If a dim2 attribute is specified, it must be of fixed type and the rank must be 2
      # Also, the shape attribute must be a list of integers with cardinality equal to the rank
      if self.child.attrs["dim2"][0] != "fixed" \
         or self.child.attrs["rank"][1] != "2" \
         or not isinstance(self.child.attrs["shape"][0], plist.List) \
         or self.child.attrs["shape"][0].datatype is not int \
         or str(self.child.attrs["shape"][0].cardinality) != self.child.attrs["rank"][1]:
        return False

    # Otherwise, the rank must be one and the shape an integer
    elif self.child.attrs["rank"][1] != "1" or self.child.attrs["shape"][0] is not int:
      return False

    # The data for the element must be a list of one or more
    if not isinstance(self.datatype, plist.List) or self.datatype.cardinality != "+":
      return False

    # If the shape has been set, check that it has a valid value
    if self.child.attrs["shape"][1] != None:
      if geometry_dim_tree.data is None:
        return False

      dim1, dim2 = self.tensor_shape(geometry_dim_tree)
      if "dim2" in self.child.attrs.keys():
        if self.child.attrs["shape"][1] != str(dim1) + " " + str(dim2):
          return False
      elif self.child.attrs["shape"][1] != str(dim1):
        return False

    return True

  def tensor_shape(self, dimension):
    """
    Read the tensor shape for tensor or vector data in the current MixedTree.
    """

    if not isinstance(dimension, int):
      dimension = int(dimension.data)

    dim1 = 1
    dim2 = 1
    if "dim1" in self.child.attrs.keys():
      dim1 = int(eval(self.child.attrs["dim1"][1], {"dim":dimension}))
      if "dim2" in self.child.attrs.keys():
        dim2 = int(eval(self.child.attrs["dim2"][1], {"dim":dimension}))

    return (dim1, dim2)

  def is_symmetric_tensor(self, geometry):
    """
    Read if the tensor data in the current MixedTree is symmetric.
    """

    dim1, dim2 = self.tensor_shape(geometry)

    return dim1 == dim2 and "symmetric" in self.child.attrs.keys() and self.child.attrs["symmetric"][1] == "true"

  def is_code(self):
    """
    Perform a series of tests on the current MixedTree, to determine if
    it is intended to be used to store code data.
    """
    try:
      lang = self.get_attr("language")
      if lang == "python":
        return True
    except:
      pass

    if self.datatype is not str:
      return False

    if "type" in self.child.attrs.keys():
      return self.child.attrs["type"][1] == "code"
    
    return False

  def get_code_language(self):
    """
    Assuming this is a code snippet return the language.
    """

    if not self.is_code():
      return "python"

    try:
      lang = self.child.get_attr("language")
      return lang
    except:
      try:
        lang = self.get_attr("language")
        return lang
      except:   
        return "python"

  def get_name_path(self, leaf = True):
    return self.parent.get_name_path(leaf)

  def is_sliceable(self):
    return True

  def all_attrs_fixed(self):
    for attr in self.attrs:
      if self.attrs[attr][0] != "fixed":
        return False

    return True



#!/usr/bin/env python

#    This file is part of Diamond.
#
#    Diamond is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Diamond is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Diamond.  If not, see <http://www.gnu.org/licenses/>.

import base64
import bz2
import copy
import sys

import cStringIO

from lxml import etree

import debug
import choice
import plist
import preprocess
import tree

def memoise(f):
    cache = {}
    def memf(*x):
        if x not in cache:
            cache[x] = f(*x)
        return cache[x]
    return memf

##########################
#     SCHEMA CLASS       #
##########################

class Schema(object):
  def __init__(self, schemafile):
    p = etree.XMLParser(remove_comments=True)
    self.tree = etree.parse(cStringIO.StringIO(preprocess.preprocess(schemafile)), p)

    self.callbacks = {'element': self.cb_element,
                      'documentation': self.cb_documentation,
                      'value': self.cb_value,
                      'attribute': self.cb_attribute,
                      'data': self.cb_data,
                      'optional': self.cb_optional,
                      'zeroOrMore': self.cb_zeroormore,
                      'oneOrMore': self.cb_oneormore,
                      'choice': self.cb_choice,
                      'empty': self.cb_empty,
                      'list': self.cb_list,
                      'group': self.cb_group,
                      'interleave': self.cb_group,
                      'name': self.cb_name,
                      'text': self.cb_text,
                      'anyName' : self.cb_anyname,
                      'nsName' : self.cb_nsname,
                      'except' : self.cb_except,
                      'ignore' : self.cb_ignore,
                      'notAllowed' : self.cb_notallowed}

    self.lost_eles  = []
    self.added_eles = []
    self.lost_attrs  = []
    self.added_attrs = []

    return

  @memoise
  def element_children(self, element):
    """
    Return a list of the children of the supplied element, following references
    as required.
    """

    children = []
    for child1 in element.iterchildren(tag=etree.Element):
      if self.tag(child1) == "ref":
        if not "name" in child1.keys():
          debug.deprint("Warning: Encountered reference with no name")
          continue

        name = child1.get("name")

        xpath = self.tree.xpath('/t:grammar/t:define[@name="' + name + '"]',
               namespaces={'t': 'http://relaxng.org/ns/structure/1.0'})

        if len(xpath) == 0:
          debug.deprint("Warning: Schema reference %s not found" % name, 0)
          continue

        for child2 in self.element_children(xpath[0]):
          children.append(child2)
      else:
        children.append(child1)

    return children
    
  def choice_children(self, children):
    """
    Collapse all choices within a choice into a single list of (non-choice) children
    """
  
    out_children = []
    for child in children:
      if self.tag(child) == "choice":
        out_children = out_children + self.choice_children(self.element_children(child))
      else:
        out_children.append(child)
        
    return out_children

  def valid_children(self, eid):
    if isinstance(eid, tree.Tree):
      eid = eid.schemaname

    if eid == ":start":
      try:
        node = self.tree.xpath('/t:grammar/t:start', namespaces={'t': 'http://relaxng.org/ns/structure/1.0'})[0]
      except:
        debug.deprint("No valid start node found. Are you using a library Relax-NG file like spud_base.rng?", 0)
        sys.exit(0)
    else:
      xpath = self.tree.xpath(eid)
      if len(xpath) == 0:
        debug.deprint("Warning: no element with XPath %s" % eid)
        return []
      node = xpath[0]

    results = []

    for child in self.element_children(node):
      self.append(results, self.to_tree(child))

    if eid == ":start" and len(results) != 1:
      debug.deprint("Error: there must be exactly one root element in an XML document, but found:", 0)
      for result in results:
        debug.deprint("  %s" % result.name, 0)
      sys.exit(1)
      
    return results

  def valid_node(self, eid):
    if isinstance(eid, tree.Tree) or isinstance(eid, choice.Choice):
      eidtree = eid
      eid = eid.schemaname

    xpath = self.tree.xpath(eid)
    if len(xpath) == 0:
      debug.deprint("Warning: no element with XPath %s" % eid)
      return None
    node = xpath[0]

    node = self.to_tree(node)
  
    if eidtree is not None:
      node.cardinality = eidtree.cardinality
      node.parent = eidtree.parent

    return node

  def to_tree(self, element):
    tag = self.tag(element)
    f = self.callbacks[tag]
    facts = {}
    x = f(element, facts)
    return x    

  #############################################
  # Beginning of schema processing functions. #
  #############################################

  def cb_name(self, element, facts):
    name = element.text
    facts["name"] = name

  def cb_element(self, element, facts):
    newfacts = {}
    if "cardinality" in facts:
      newfacts["cardinality"] = facts["cardinality"]

    if "name" in element.keys():
      newfacts["name"] = element.get("name")
    else:
      debug.deprint("Warning: Encountered element with no name")

    newfacts['schemaname'] = self.tree.getpath(element)

    for child in self.element_children(element):
      tag = self.tag(child)

      if tag not in ['element', 'optional', 'zeroOrMore', 'oneOrMore', 'ignore']:
        f = self.callbacks[tag]
        x = f(child, newfacts)

    try:
      d = newfacts["datatype"]
      if isinstance(d, tuple):
        new_d = []

        for x in d:
          if x is not None:
            new_d.append(x)

        d = tuple(new_d)
        newfacts["datatype"] = d
        if len(d) == 0:
          newfacts["datatype"] = None
        elif len(d) == 1 and isinstance(d[0], plist.List):
          newfacts["datatype"] = d[0]
        else:
          l_values = []
          l_data   = []
          for x in d:
            if isinstance(x, str):
              l_values.append(x)
            else:
              l_data.append(x)

          if len(l_data) > 1:
            if "name" in element.keys():
              debug.deprint("Warning: Element %s has multiple datatypes - using first one" % newfacts["name"])
            else:
              debug.deprint("Warning: Unnamed element has multiple datatypes - using first one")

          if len(l_data) > 0:
            if len(l_values) == 0:
              newfacts["datatype"] = l_data[0]
            else:
              newfacts["datatype"] = tuple([tuple(l_values)] + l_data[0])
    except KeyError:
      pass

    return tree.Tree(**newfacts)

  def cb_documentation(self, element, facts):
    facts['doc'] = element.text

  def cb_value(self, element, facts):
    if "datatype" in facts:
      l = list(facts["datatype"])
    else:
      l = []

    l.append(element.text)
    facts["datatype"] = tuple(l)

  def cb_attribute(self, element, facts):
    if not "name" in element.keys():
      debug.deprint("Warning: Encountered attribute with no name")
      return

    newfacts = {}
    name = element.get("name")

    for child in self.element_children(element):
      tag = self.tag(child)
      f = self.callbacks[tag]
      x = f(child, newfacts)

    if "attrs" not in facts:
      facts["attrs"] = {}

    try:
      datatype = newfacts["datatype"]
    except:
      debug.deprint("Warning: Encountered attribute with no datatype")
      return
    curval = None

    if isinstance(datatype, tuple):
      new_datatype = []
      for x in datatype:
        if not x is None:
          new_datatype.append(x)
      datatype = new_datatype
      if len(datatype) == 0:
        datatype = None
      elif len(datatype) == 1:
        datatype = datatype[0]
        if isinstance(datatype, str):
          curval = datatype
          datatype = 'fixed'
        else:
          curval = None
      else:
        l_values = []
        l_data   = []
        for x in datatype:
          if isinstance(x, str):
            l_values.append(x)
          else:
            l_data.append(x)

        if len(l_data) > 0:
          debug.deprint("Warning: Attribute %s has multiple datatypes - using first one" % name)
          if len(l_values) == 0:
            datatype = l_data[0]
          else:
            datatype = tuple([tuple(l_values)] + l_data[0])
        else:
          datatype = tuple(l_values)

    facts["attrs"][name] = (datatype, curval)

  def cb_data(self, element, facts):
    if "datatype" in facts:
      if isinstance(facts["datatype"], tuple):
        l = list(facts["datatype"])
      else:
        l = [facts["datatype"]]
    else:
      l = []

    mapping = {'integer': int,
               'float': float,
               'double': float,
               'decimal': float,
               'string': str,
               'ID' : str,
               'anyURI' : str,
               'IDREF' : int,
               'NMTOKEN' : str,
               'boolean': bool}

    datatype_name = element.get("type")
    l.append(mapping[datatype_name])
    if len(l) == 1:
      facts["datatype"] = l[0]
    else:
      facts["datatype"] = tuple(l)

  def cb_optional(self, element, facts):
    facts["cardinality"] = '?'
    r = []
    for child in self.element_children(element):
      tag = self.tag(child)
      f = self.callbacks[tag]
      self.append(r, f(child, facts))

    return r

  def cb_zeroormore(self, element, facts):
    facts["cardinality"] = '*'
    r = []
    for child in self.element_children(element):
      tag = self.tag(child)
      f = self.callbacks[tag]
      self.append(r, f(child, facts))

    return r

  def cb_oneormore(self, element, facts):
    facts["cardinality"] = '+'
    r = []
    for child in self.element_children(element):
      tag = self.tag(child)
      f = self.callbacks[tag]
      self.append(r, f(child, facts))

    return r

  def cb_choice(self, element, facts):
    # there are really two cases here.
    # choice between values of elements,
    # and choice between elements

    tagnames = [self.tag(child) for child in element]

    if "value" in tagnames:
      for child in self.element_children(element):
        tag = self.tag(child)
        f = self.callbacks[tag]
        x = f(child, facts)

    else:
      if "schemaname" in facts:
        return

      facts['schemaname'] = self.tree.getpath(element)

      r = []
      children = self.choice_children(self.element_children(element))
      
      # bloody simplified RNG
      if len(children) == 2:
        empty = [x for x in children if self.tag(x) == "empty"]
        if empty:
          nonempty = [x for x in children if self.tag(x) != "empty"]
          tag = self.tag(nonempty[0])
          f = self.callbacks[tag]
          return f(element, facts)

      for child in children:
        newfacts = {}
        tag = self.tag(child)
        f = self.callbacks[tag]
        self.append(r, f(child, newfacts))

      return choice.Choice(r, **facts)

  def cb_empty(self, element, facts):
    pass

  def cb_list(self, element, facts):
    newfacts = {}
    for child in self.element_children(element):
      tag = self.tag(child)
      f = self.callbacks[tag]
      f(child, newfacts)

    d = newfacts["datatype"]
    try:
      c = newfacts["cardinality"]
    except KeyError:
      c = ''
      if isinstance(d, tuple):
        c = str(len(d))
        d = d[0]

    l = plist.List(d, c)
    if "datatype" in facts:
      e = list(facts["datatype"])
    else:
      e = []

    e.append(l)
    facts["datatype"] = tuple(e)

  def cb_group(self, element, facts):
    results = []
    for child in self.element_children(element):
      newfacts = {}
      tag = self.tag(child)
      f = self.callbacks[tag]
      self.append(results, f(child, newfacts))

    return results

  def cb_text(self, element, facts):
    if "datatype" in facts:
      if isinstance(facts["datatype"], tuple):
        l = list(facts["datatype"])
      else:
        l = [facts["datatype"]]
    else:
      l = []

    l.append(str)
    if len(l) == 1:
      facts["datatype"] = l[0]
    else:
      facts["datatype"] = tuple(l)

  def cb_anyname(self, element, facts):
    debug.deprint("anyName element found. Yet to handle.", 0)
#    sys.exit(1)

  def cb_nsname(self, element, facts):
    debug.deprint("nsName element found. Yet to handle.", 0)
#    sys.exit(1)

  def cb_except(self, element, facts):
    debug.deprint("except element found. Yet to handle.", 0)
#    sys.exit(1)

  def cb_ignore(self, element, facts):
    pass

  def cb_notallowed(self, element, facts):
    debug.dprint("notallowed element found. Yet to handle.", 0)

  #######################################
  # End of schema processing functions. #
  #######################################

  def tag(self, element):
    # Ignore non-RelaxNG elements. Is this the best way to handle it? 
    namespace = element.tag.split('}')[0]
    if namespace.find("relaxng") != -1:
      return element.tag.split('}')[-1]
    else:
      return "ignore"

  # append - append either a list or single element 'x' to 'r'.
  def append(self, r, x):
    if x is None:
      return
    
    if isinstance(x, list):
      r.extend(x)
    else:
      r.append(x)

  ##########################################
  # Beginning of XML processing functions. #
  ##########################################

  # read takes a file handle, constructs a generic in-memory representation using the
  # the etree API, and then converts it to a tree of Tree and Choice elements.
  def read(self, xmlfile, root = None):
    try:
      doc = etree.parse(xmlfile)
    except etree.XMLSyntaxError as e:
      debug.dprint("Invalid XML.")
      debug.dprint(e)
      return None

    self.lost_eles = []
    self.added_eles = []
    self.lost_attrs  = []
    self.added_attrs = []

    if root is None:
      datatree = self.valid_children(":start")[0]
    else:
      datatree = self.valid_node(root)

    xmlnode  = doc.getroot()
    self.xml_read_merge(datatree, xmlnode)
    self.xml_read_core(datatree.get_current_tree(), xmlnode, doc)

    if len(self.lost_eles) != 0:
      debug.deprint("WARNING: Lost XML elements:\n" + str(self.lost_eles))
    if len(self.added_eles) != 0:
      debug.deprint("WARNING: Added XML elements:\n" + str(self.added_eles))
    if len(self.lost_attrs) != 0:
      debug.deprint("WARNING: Lost XML attributes:\n" + str(self.lost_attrs))
    if len(self.added_eles) != 0:
      debug.deprint("WARNING: Added XML attributes:\n" + str(self.added_attrs))
      
    return datatree

  def xml_read_merge(self, datatree, xmlnode):
    # The datatree has the following set:
    # name, schemaname, doc, cardinality, datatype, parent,
    # attribute datatypes.
    # the xmlnode contains the following information:
    # attribute values, data
    # merge the two.
    
    datatree.xmlnode = xmlnode
    xmlkeys = xmlnode.keys()

    if datatree.__class__ is tree.Tree:
      to_set = datatree
    elif datatree.__class__ is choice.Choice:
      if "name" in xmlkeys:
        xmlname = xmlnode.get("name")
        have_found = False

        possibles = [tree_choice for tree_choice in datatree.get_choices() if tree_choice.name == xmlnode.tag]
        # first loop over the fixed-value names
        for tree_choice in possibles:
          if "name" not in tree_choice.attrs:
            continue

          datatype = tree_choice.attrs["name"][0]
          if datatype == 'fixed':
            treename = tree_choice.attrs["name"][1]
            if treename == xmlname:
              have_found = True
              datatree.set_active_choice_by_ref(tree_choice)
              break

        # if we haven't found it, look for a generic name
        if have_found is False:
          for tree_choice in possibles:
            if "name" not in tree_choice.attrs:
              continue

            datatype = tree_choice.attrs["name"][0]
            if datatype != 'fixed':
              have_found = True
              datatree.set_active_choice_by_ref(tree_choice)
              break

      else:
        datatree.set_active_choice_by_name(xmlnode.tag)

      to_set = datatree.get_current_tree()

    # catch any lost XML attributes
    for key in xmlkeys:
      if key not in to_set.attrs.keys():
        self.lost_attrs += [to_set.name + '/' + key]

    # attribute values.
    for key in to_set.attrs.keys():
      if key in xmlkeys:
        try:
          to_set.set_attr(key, xmlnode.get(key))
        except:
          pass
      else:
        self.added_attrs += [to_set.name + '/' + key]

    # Get the text value (the node's data)
    if xmlnode.text is not None:
     try:
       text=xmlnode.text.strip()
       if text != "":
         to_set.set_data(text)
     except:
       pass
      

    # data.
    for child in xmlnode.iterchildren(tag=etree.Element):
      if child.tail is not None:
         try:
           text = child.tail.strip()
           if text != "":
             to_set.set_data(text)
             break
         except:
           pass

    to_set.recompute_validity()
    datatree.recompute_validity()

  ###########################################################################################
  # construct the priority queue
  # we treat compulsory nodes first, then descend through the cardinalities
  ###########################################################################################  
  
  def construct_priority_queue(self, schemachildren):
    # priority_queue will store the schemachildren, in the order in which
    # they query data from the XML
    priority_queue = []
    
    # compulsory first.
    for schemachild in schemachildren:
      if schemachild.cardinality == '':
        priority_queue.append(schemachild)

    # then oneormore.
    for schemachild in schemachildren:
      if schemachild.cardinality == '+':
        priority_queue.append(schemachild)

    # then, optional
    for schemachild in schemachildren:
      if schemachild.cardinality == '?':
        priority_queue.append(schemachild)

    # then zeroormore.
    for schemachild in schemachildren:
      if schemachild.cardinality == '*':
        priority_queue.append(schemachild)
        
    return priority_queue

  ###########################################################################################
  # initialise the availability data
  # avail[name][xmlnode] records whether xmlnode is available or not
  ###########################################################################################
  def init_avail_data(self, xmlnode, schemachildren):
    used = {}
    avail = {}
    
    for xml in xmlnode.iterchildren(tag=etree.Element):
      used[xml] = False
    
    for schemachild in schemachildren:
      for name in schemachild.get_possible_names():
        avail[name] = {}

    for name in avail:
      for xmldata in xmlnode.iterchildren(tag=name):
        avail[name][xmldata] = True
        
    return (used, avail)

  ###########################################################################################
  # assign the available xml nodes to the children the schema says should be there
  # in order of priority.
  # xmls[schemachild.schemaname] is the list of xml nodes
  # that schemachild should take
  ###########################################################################################
  def assign_xml_nodes(self, priority_queue, xmlnode, avail):   
    xmls = {}

    for schemachild in priority_queue:
      if schemachild.cardinality in ['', '?']:
        for curtree in schemachild.get_choices():
          name = curtree.name

          have_fixed_name = False
          if "name" in curtree.attrs:
            datatype = curtree.attrs["name"][0]
            if datatype == 'fixed':
              have_fixed_name = True

          if have_fixed_name is False:
            xml = xmlnode.xpath(name)
          else:
            xml = xmlnode.xpath(name + '[@name="%s"]' % curtree.get_attr("name"))

          for xmldata in xml:
            if avail[name][xmldata]:
              avail[name][xmldata] = False
              xmls[schemachild.schemaname] = [xmldata]
              break 

          if schemachild.schemaname not in xmls:
            if schemachild.cardinality == '':
              xmls[schemachild.schemaname] = copy.deepcopy([])
            else:
              xmls[schemachild.schemaname] = copy.deepcopy([])
      elif schemachild.cardinality in ['*', '+']:
        xmls[schemachild.schemaname] = copy.deepcopy([])
        for curtree in schemachild.get_choices():
          name = curtree.name

          have_fixed_name = False
          if "name" in curtree.attrs:
            datatype = curtree.attrs["name"][0]
            if datatype == 'fixed':
              have_fixed_name = True

          if have_fixed_name is False:
            xml = xmlnode.xpath(name)
          else:
            xml = xmlnode.xpath(name + '[@name="%s"]' % curtree.get_attr("name"))

          for xmldata in xml:
            if avail[name][xmldata]:
              avail[name][xmldata] = False
              xmls[schemachild.schemaname].append(xmldata)
                
    return xmls

  ###########################################################################################
  # now that we have assigned the xml nodes, loop through and grab them
  # stuff the tree data in bins[schemachild.schemaname]
  ###########################################################################################
  def assign_xml_children(self, priority_queue, xmlnode, xmls, schemachildren, used, rootdoc):
    # bins[schemachild.schemaname] will store the data associated with schemachild
    bins = {}

    for schemachild in schemachildren:
      bins[schemachild.schemaname] = []
      
    for schemachild in priority_queue:
      if schemachild.cardinality in ['', '?']:
        child = schemachild.copy()
        child.xmlnode = None
        child.active = True
        if len(xmls[schemachild.schemaname]) == 1:
          xmldata = xmls[schemachild.schemaname][0]
          used[xmldata] = True
          self.xml_read_merge(child, xmldata)

          # Was this part of the uncompressed XML file or part of a hidden comment?
          if xmldata.getroottree().getroot() != rootdoc.getroot():
            self.xml_read_core(child.get_current_tree(), xmldata, xmldata.getroottree())
            child.active = False
            child.recurse = False
        else:
          if schemachild.cardinality == '?':
            child.active = False

        bins[schemachild.schemaname] = [child]

      elif schemachild.cardinality in ['*', '+']:
        for xmldata in xmls[schemachild.schemaname]:
          child = schemachild.copy()
          child.active = True
          used[xmldata] = True
          self.xml_read_merge(child, xmldata)
          bins[schemachild.schemaname].append(child)

      if schemachild.cardinality == '+':
        # check that we have at least one.
        count = len(bins[schemachild.schemaname])
        if count == 0:
          child = schemachild.copy()
          child.active = True
          child.xmlnode = None
          bins[schemachild.schemaname] = [child]

      if schemachild.cardinality in ['*', '+']:
        # add an inactive instance
        child = schemachild.copy()
        child.active = False
        child.xmlnode = None
        bins[schemachild.schemaname].append(child)

      # search for neglected choices
      if isinstance(schemachild, choice.Choice) and schemachild.cardinality in ['', '?']:
        for child in bins[schemachild.schemaname]:

          # Does the child have a valid XML node attached?
          if not hasattr(child, "xmlnode"): continue
          if child.xmlnode is None: continue

          current_choice = child.get_current_tree()
          for tree_choice in child.get_choices():
            if tree_choice is current_choice: continue

    return bins

  # Loop over lost nodes, and store their XML so the user can be notified later.
  def check_unused_nodes(self, used): 
    def xml2string(xml):
      buf = cStringIO.StringIO()
      buf.write(etree.tostring(xml, pretty_print = True))
      s = buf.getvalue()
      buf.close()
      
      return s
  
    for xml in used:
      if used[xml] is False:
        s = xml2string(xml)
        self.lost_eles += [s]
  
  # Append the children to the datatree in the order the schema presents them.
  # Order matters here.
  def append_children(self, schemachildren, datatree, bins):
    for schemachild in schemachildren:
      for child in bins[schemachild.schemaname]:
        child.set_parent(datatree)
        datatree.children.append(child)  
      
  # Recurse down the in-memory XML tree, reading elements and merging their
  # information into the in-memory Tree structure.
  def read_children(self, datatree, rootdoc):
    for schild in datatree.children:

      if hasattr(schild, "recurse"):
        if schild.recurse is False:
          continue

      if schild.__class__ is choice.Choice:
        child = schild.get_current_tree()
      else:
        child = schild

      if schild.active is False:
        continue

      child.children = copy.copy([])
      self.xml_read_core(child, schild.xmlnode, rootdoc)

  # xml_read_core recurses throughout the tree, calling xml_read_merge on the current node "xmlnode" and
  # and reading information about the node's children.
  def xml_read_core(self, datatree, xmlnode, rootdoc):
    """This is the part that recurses, you see."""

    assert len(datatree.children) == 0

    # no information from XML to be had :-/
    if xmlnode is None:
      self.added_eles.append(self.readable_name(datatree))
      if datatree.active: datatree.add_children(self)
      return

    schemachildren = self.valid_children(datatree)
    
    priority_queue = self.construct_priority_queue(schemachildren) 
    (used, avail) = self.init_avail_data(xmlnode, schemachildren)
    xmls = self.assign_xml_nodes(priority_queue, xmlnode, avail)
    bins = self.assign_xml_children(priority_queue, xmlnode, xmls, schemachildren, used, rootdoc)
    self.append_children(schemachildren, datatree, bins)              
    self.check_unused_nodes(used)
    self.read_children(datatree, rootdoc)

    datatree.recompute_validity()
    
  def read_errors(self):
    return self.lost_eles, self.added_eles, self.lost_attrs, self.added_attrs

  def readable_name(self, datatree):
    output = ""
    node = datatree
    while node is not None:
      output = node.name + '/' + output
      node = node.parent
    return '/' + output[:-1]



#!/usr/bin/env python

#    This file is part of Diamond.
#
#    Diamond is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Diamond is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Diamond.  If not, see <http://www.gnu.org/licenses/>.

import plist

def print_type(datatype, bracket = True):
  """
  Create a string to be displayed in place of empty data / attributes.
  """

  def familiar_type(type_as_printable):
    """
    Convert some type names to more familiar equivalents.
    """

    if type_as_printable == "decim":
      return "float"
    elif type_as_printable == "int":
      return "integer"
    elif type_as_printable == "str":
      return "string"
    else:
      return type_as_printable

  def type_name(datatype):
    """
    Return a human readable version of datatype.
    """

    datatype_string = str(datatype)

    if datatype_string[:7] == "<type '" and datatype_string[len(datatype_string) - 2:] == "'>":
      value_type_split = datatype_string.split("'")
      return familiar_type(value_type_split[1])

    value_type_split1 = datatype_string.split(".")
    value_type_split2 = value_type_split1[len(value_type_split1) - 1].split(" ")
    if len(value_type_split2) == 1:
      return familiar_type(value_type_split2[0][0:len(value_type_split2[0]) - 6])
    else:
      return familiar_type(value_type_split2[0])

  #print_type

  if isinstance(datatype, plist.List):
    if (isinstance(datatype.cardinality, int) and datatype.cardinality == 1) or datatype.cardinality == "":
      type_as_printable = type_name(datatype.datatype).lower()
    else:
      type_as_printable = type_name(datatype).lower() + " of "
      list_type_as_printable = type_name(datatype.datatype).lower()
      if isinstance(datatype.cardinality, int):
        type_as_printable += str(datatype.cardinality) + " " + list_type_as_printable + "s"
      else:
        type_as_printable += list_type_as_printable + "s"
  else:
    type_as_printable = type_name(datatype).lower()

  if bracket:
    type_as_printable = "(" + type_as_printable + ")"

  return type_as_printable



#!/usr/bin/env python

#    This file is part of Diamond.
#
#    Diamond is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Diamond is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Diamond.  If not, see <http://www.gnu.org/licenses/>.

import gobject
import gtk
import os
import threading

import schemausage

RELAXNGNS = "http://relaxng.org/ns/structure/1.0"
RELAXNG = "{" + RELAXNGNS + "}"

class UseView(gtk.Window):
  def __init__(self, schema, suffix, folder = None):
    gtk.Window.__init__(self)
    self.__add_controls()

    if folder is None:
      dialog = gtk.FileChooserDialog(title = "Input directory",
                                   action = gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER,
                                   buttons = (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL, gtk.STOCK_OPEN, gtk.RESPONSE_OK))

      response = dialog.run()
      if response != gtk.RESPONSE_OK:
        dialog.destroy()
        return

      folder = os.path.abspath(dialog.get_filename())
      dialog.destroy()
    #endif

    paths = []
    for dirpath, dirnames, filenames in os.walk(folder):
      paths.extend([os.path.join(dirpath, filename) for filename in filenames if filename.endswith(suffix)])

    self.__update(schema, paths)
    self.show_all()

  def __add_controls(self): 
    self.set_title("Unused schema entries")
    self.set_default_size(800, 600)

    vbox = gtk.VBox()

    scrolledwindow = gtk.ScrolledWindow()
    scrolledwindow.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)

    self.treeview = gtk.TreeView()

    self.treeview.get_selection().set_mode(gtk.SELECTION_SINGLE)

    # Node column
    celltext = gtk.CellRendererText()
    column = gtk.TreeViewColumn("Node", celltext)
    column.set_cell_data_func(celltext, self.set_celltext)

    self.treeview.append_column(column)

    # 0: The node tag
    # 1: Used (0 == Not used, 1 = Child not used, 2 = Used)
    self.treestore = gtk.TreeStore(gobject.TYPE_PYOBJECT, gobject.TYPE_PYOBJECT)
    self.treeview.set_enable_search(False)

    scrolledwindow.add(self.treeview)
    vbox.pack_start(scrolledwindow)

    self.statusbar = gtk.Statusbar()
    vbox.pack_end(self.statusbar, expand = False)
    self.add(vbox)

  def __set_treestore(self, node):
    def set_treestore(node, iter, type):
      if node.tag == RELAXNG + "element":
        name = schemausage.node_name(node)
        if name == "comment":
          return #early out to skip comment nodes

        tag = name + (type if type else "")
        child_iter = self.treestore.append(iter, [tag, 2])
        self.mapping[self.tree.getpath(node)] = self.treestore.get_path(child_iter)
        type = None
      elif node.tag == RELAXNG + "choice" and all(n.tag != RELAXNG + "value" for n in node):
        tag = "choice" + (type if type else "")
        child_iter = self.treestore.append(iter, [tag, 2])
        self.mapping[self.tree.getpath(node)] = self.treestore.get_path(child_iter)
        type = None
      elif node.tag == RELAXNG + "optional":
        child_iter = iter
        type = " ?"
      elif node.tag == RELAXNG + "oneOrMore":
        child_iter = iter
        type = " +"
      elif node.tag == RELAXNG + "zeroOrMore":
        child_iter = iter
        type = " *"
      elif node.tag == RELAXNG + "ref":
        query = '/t:grammar/t:define[@name="' + node.get("name") + '"]'
        if query not in cache:
          cache[query] = self.tree.xpath(query, namespaces={'t': RELAXNGNS})[0]
        node = cache[query]
        child_iter = iter
      elif node.tag == RELAXNG + "group" or node.tag == RELAXNG + "interleave":
        child_iter = iter
      else:
        return

      for child in node:
        set_treestore(child, child_iter, type)

    cache = {}
    set_treestore(node, None, None)

  def __set_useage(self, useage):
    for xpath in useage:
      try:
        iter = self.treestore.get_iter(self.mapping[xpath])
        self.treestore.set_value(iter, 1, 0)
      except KeyError:
        pass #probably a comment node

  def __floodfill(self, iter, parent = 2):
    """
    Floodfill the tree with the correct useage.
    """
    if parent == 0: #parent is not used
      self.treestore.set_value(iter, 1, 0) #color us not used

    useage = self.treestore.get_value(iter, 1)

    child = self.treestore.iter_children(iter)
    while child is not None:
      change = self.__floodfill(child, useage)
      if change != 2 and useage == 2:
        self.treestore.set(iter, 1, 1)
      child = self.treestore.iter_next(child)

    return self.treestore.get_value(iter, 1)
 

  def __update(self, schema, paths):
    self.tree = schema.tree
    start = self.tree.xpath('/t:grammar/t:start', namespaces={'t': RELAXNGNS})[0]
    self.mapping = {}

    def async_update(self, start, schema, paths, context):
      gtk.idle_add(self.statusbar.push, context, "Parsing schema")
      self.__set_treestore(start[0])
      gtk.idle_add(self.statusbar.push, context, "Schema parsed... finding usage")
      self.__set_useage(schemausage.find_unusedset(schema, paths))
      gtk.idle_add(self.statusbar.push, context, "Usage found")
      self.__floodfill(self.treestore.get_iter_root())
      gtk.idle_add(self.statusbar.push, context, "")
      gtk.idle_add(self.treeview.set_model, self.treestore)

    t = threading.Thread(target = async_update, args = (self, start, schema, paths, self.statusbar.get_context_id("update")))
    t.start()

  def set_celltext(self, column, cell, model, iter):
    tag, useage = model.get(iter, 0, 1)
    cell.set_property("text", tag)

    if useage == 0:
      cell.set_property("foreground", "red")
    elif useage == 1:
      cell.set_property("foreground", "indianred")
    else:
      cell.set_property("foreground", "black")



### Copyright (C) 2005 Thomas M. Hinkle
### Copyright (C) 2007 Imperial College London and others.
### Please see the AUTHORS file for a full list of contributors.
###
### This library is free software; you can redistribute it and/or
### modify it under the terms of the GNU General Public License as
### published by the Free Software Foundation; either version 2 of the
### License, or (at your option) any later version.
###
### This library is distributed in the hope that it will be useful,
### but WITHOUT ANY WARRANTY; without even the implied warranty of
### MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
### General Public License for more details.
###
### You should have received a copy of the GNU General Public License
### along with this library; if not, write to the Free Software
### Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
### 02111-1307, USA.

import pango,gtk, xml.sax.saxutils

import debug

class PangoBuffer (gtk.TextBuffer):
    desc_to_attr_table = {
        'family':[pango.AttrFamily,""],
        'style':[pango.AttrStyle,pango.STYLE_NORMAL],
        'variant':[pango.AttrVariant,pango.VARIANT_NORMAL],
        'weight':[pango.AttrWeight,pango.WEIGHT_NORMAL],
        'stretch':[pango.AttrStretch,pango.STRETCH_NORMAL],
        }
    pango_translation_properties={
            # pango ATTR TYPE : (pango attr property / tag property)
            pango.ATTR_SIZE : 'size',
            pango.ATTR_WEIGHT: 'weight',
            pango.ATTR_UNDERLINE: 'underline',
            pango.ATTR_STRETCH: 'stretch',
            pango.ATTR_VARIANT: 'variant',
            pango.ATTR_STYLE: 'style',
            pango.ATTR_SCALE: 'scale',
            pango.ATTR_STRIKETHROUGH: 'strikethrough',
            pango.ATTR_RISE: 'rise',
            }
    attval_to_markup={
            'underline':{pango.UNDERLINE_SINGLE:'single',
                         pango.UNDERLINE_DOUBLE:'double',
                         pango.UNDERLINE_LOW:'low',
                         pango.UNDERLINE_NONE:'none'},
            'stretch':{pango.STRETCH_ULTRA_EXPANDED:'ultraexpanded',
                       pango.STRETCH_EXPANDED:'expanded',
                       pango.STRETCH_EXTRA_EXPANDED:'extraexpanded',
                       pango.STRETCH_EXTRA_CONDENSED:'extracondensed',
                       pango.STRETCH_ULTRA_CONDENSED:'ultracondensed',
                       pango.STRETCH_CONDENSED:'condensed',
                       pango.STRETCH_NORMAL:'normal',
                       },
            'variant':{pango.VARIANT_NORMAL:'normal',
                       pango.VARIANT_SMALL_CAPS:'smallcaps',
                       },
            'style':{pango.STYLE_NORMAL:'normal',
                     pango.STYLE_OBLIQUE:'oblique',
                     pango.STYLE_ITALIC:'italic',
                     },
            'stikethrough':{1:'true',
                            True:'true',
                            0:'false',
                            False:'false'},
            }
    def __init__ (self):
        self.tagdict = {}
        self.tags = {}
        #self.buf = buf
        #self.set_text(txt)
        gtk.TextBuffer.__init__(self)

    def set_text (self, txt):
        gtk.TextBuffer.set_text(self,"")
        try:
            self.parsed,self.txt,self.separator = pango.parse_markup(txt,u'\x00')
        except:
            debug.deprint('Escaping text, we seem to have a problem here!', 2)
            txt=xml.sax.saxutils.escape(txt)
            self.parsed,self.txt,self.separator = pango.parse_markup(txt,u'\x00')
        self.attrIter = self.parsed.get_iterator()
        self.add_iter_to_buffer()
        while self.attrIter.next():
            self.add_iter_to_buffer()

    def add_iter_to_buffer (self):
        range=self.attrIter.range()
        font,lang,attrs = self.attrIter.get_font()
        tags = self.get_tags_from_attrs(font,lang,attrs)
        text = self.txt[range[0]:range[1]]
        if tags: self.insert_with_tags(self.get_end_iter(),text,*tags)
        else: self.insert_with_tags(self.get_end_iter(),text)

    def get_tags_from_attrs (self, font,lang,attrs):
        tags = []
        if font:
            font,fontattrs = self.fontdesc_to_attrs(font)
            fontdesc = font.to_string()
            if fontattrs:
                attrs.extend(fontattrs)
            if fontdesc and fontdesc!='Normal':
                if not self.tags.has_key(font.to_string()):
                    tag=self.create_tag()
                    tag.set_property('font-desc',font)
                    if not self.tagdict.has_key(tag): self.tagdict[tag]={}
                    self.tagdict[tag]['font_desc']=font.to_string()
                    self.tags[font.to_string()]=tag
                tags.append(self.tags[font.to_string()])
        if lang:
            if not self.tags.has_key(lang):
                tag = self.create_tag()
                tag.set_property('language',lang)
                self.tags[lang]=tag
            tags.append(self.tags[lang])
        if attrs:
            for a in attrs:
                if a.type == pango.ATTR_FOREGROUND:
                    gdkcolor = self.pango_color_to_gdk(a.color)
                    key = 'foreground%s'%self.color_to_hex(gdkcolor)
                    if not self.tags.has_key(key):
                        self.tags[key]=self.create_tag()
                        self.tags[key].set_property('foreground-gdk',gdkcolor)
                        self.tagdict[self.tags[key]]={}
                        self.tagdict[self.tags[key]]['foreground']="#%s"%self.color_to_hex(gdkcolor)
                    tags.append(self.tags[key])
                if a.type == pango.ATTR_BACKGROUND:
                    gdkcolor = self.pango_color_to_gdk(a.color)
                    key = 'background%s'%self.color_to_hex(gdkcolor)
                    if not self.tags.has_key(key):
                        self.tags[key]=self.create_tag()
                        self.tags[key].set_property('background-gdk',gdkcolor)
                        self.tagdict[self.tags[key]]={}
                        self.tagdict[self.tags[key]]['background']="#%s"%self.color_to_hex(gdkcolor)
                    tags.append(self.tags[key])
                if self.pango_translation_properties.has_key(a.type):
                    prop=self.pango_translation_properties[a.type]
                    #print 'setting property %s of %s (type: %s)'%(prop,a,a.type)
                    val=getattr(a,'value')
                    #tag.set_property(prop,val)
                    mval = val
                    if self.attval_to_markup.has_key(prop):
                        #print 'converting ',prop,' in ',val
                        if self.attval_to_markup[prop].has_key(val):
                            mval = self.attval_to_markup[prop][val]
                        else:
                            debug.deprint("hmmm, didn't know what to do with value %s"%val, 2)
                    key="%s%s"%(prop,val)
                    if not self.tags.has_key(key):
                        self.tags[key]=self.create_tag()
                        self.tags[key].set_property(prop,val)
                        self.tagdict[self.tags[key]]={}
                        self.tagdict[self.tags[key]][prop]=mval
                    tags.append(self.tags[key])
        return tags

    def get_tags (self):
        tagdict = {}
        for pos in range(self.get_char_count()):
            iter=self.get_iter_at_offset(pos)
            for tag in iter.get_tags():
                if tagdict.has_key(tag):
                    if tagdict[tag][-1][1] == pos - 1:
                        tagdict[tag][-1] = (tagdict[tag][-1][0],pos)
                    else:
                        tagdict[tag].append((pos,pos))
                else:
                    tagdict[tag]=[(pos,pos)]
        return tagdict

    def get_text (self, start=None, end=None, include_hidden_chars=True):
        tagdict=self.get_tags()
        if not start: start=self.get_start_iter()
        if not end: end=self.get_end_iter()
        txt = unicode(gtk.TextBuffer.get_text(self,start,end))
        cuts = {}
        for k,v in tagdict.items():
            stag,etag = self.tag_to_markup(k)
            for st,e in v:
                if cuts.has_key(st): cuts[st].append(stag) #add start tags second
                else: cuts[st]=[stag]
                if cuts.has_key(e+1): cuts[e+1]=[etag]+cuts[e+1] #add end tags first
                else: cuts[e+1]=[etag]
        last_pos = 0
        outbuff = ""
        cut_indices = cuts.keys()
        cut_indices.sort()
        soffset = start.get_offset()
        eoffset = end.get_offset()
        cut_indices = filter(lambda i: eoffset >= i >= soffset, cut_indices)
        for c in cut_indices:
            if not last_pos==c:
                outbuff += xml.sax.saxutils.escape(txt[last_pos:c])
                last_pos = c
            for tag in cuts[c]:
                outbuff += tag
        outbuff += xml.sax.saxutils.escape(txt[last_pos:])
        return outbuff

    def tag_to_markup (self, tag):
        stag = "<span"
        for k,v in self.tagdict[tag].items():
            stag += ' %s="%s"'%(k,v)
        stag += ">"
        return stag,"</span>"

    def fontdesc_to_attrs (self,font):
        nicks = font.get_set_fields().value_nicks
        attrs = []
        for n in nicks:
            if self.desc_to_attr_table.has_key(n):
                Attr,norm = self.desc_to_attr_table[n]
                # create an attribute with our current value
                attrs.append(Attr(getattr(font,'get_%s'%n)()))
                # unset our font's value
                getattr(font,'set_%s'%n)(norm)
        return font,attrs

    def pango_color_to_gdk (self, pc):
        return gtk.gdk.Color(pc.red,pc.green,pc.blue)

    def color_to_hex (self, color):
        hexstring = ""
        for col in 'red','green','blue':
            hexfrag = hex(getattr(color,col)/(16*16)).split("x")[1]
            if len(hexfrag)<2: hexfrag = "0" + hexfrag
            hexstring += hexfrag
        return hexstring

    def apply_font_and_attrs (self, font, attrs):
        tags = self.get_tags_from_attrs(font,None,attrs)
        for t in tags: self.apply_tag_to_selection(t)

    def remove_font_and_attrs (self, font, attrs):
        tags = self.get_tags_from_attrs(font,None,attrs)
        for t in tags: self.remove_tag_from_selection(t)

    def setup_default_tags (self):
        self.italics = self.get_tags_from_attrs(None,None,[pango.AttrStyle('italic')])[0]
        self.bold = self.get_tags_from_attrs(None,None,[pango.AttrWeight('bold')])[0]
        self.underline = self.get_tags_from_attrs(None,None,[pango.AttrUnderline('single')])[0]

    def get_selection (self):
        bounds = self.get_selection_bounds()
        if not bounds:
            iter=self.get_iter_at_mark(self.get_insert())
            if iter.inside_word():
                start_pos = iter.get_offset()
                iter.forward_word_end()
                word_end = iter.get_offset()
                iter.backward_word_start()
                word_start = iter.get_offset()
                iter.set_offset(start_pos)
                bounds = (self.get_iter_at_offset(word_start),
                          self.get_iter_at_offset(word_end+1))
            else:
                bounds = (iter,self.get_iter_at_offset(iter.get_offset()+1))
        return bounds

    def apply_tag_to_selection (self, tag):
        selection = self.get_selection()
        if selection:
            self.apply_tag(tag,*selection)

    def remove_tag_from_selection (self, tag):
        selection = self.get_selection()
        if selection:
            self.remove_tag(tag,*selection)

    def remove_all_tags (self):
        selection = self.get_selection()
        if selection:
            for t in self.tags.values():
                self.remove_tag(t,*selection)

class InteractivePangoBuffer (PangoBuffer):
    def __init__ (self,
                  normal_button=None,
                  toggle_widget_alist=[]):
        """An interactive interface to allow marking up a gtk.TextBuffer.
        txt is initial text, with markup.
        buf is the gtk.TextBuffer
        normal_button is a widget whose clicked signal will make us normal
        toggle_widget_alist is a list that looks like this:
        [(widget, (font,attr)),
         (widget2, (font,attr))]
         """
        PangoBuffer.__init__(self)
        if normal_button: normal_button.connect('clicked',lambda *args: self.remove_all_tags())
        self.tag_widgets = {}
        self.internal_toggle = False
        self.insert = self.get_insert()
        self.connect('mark-set',self._mark_set_cb)
        self.connect('changed',self._changed_cb)
        for w,tup in toggle_widget_alist:
            self.setup_widget(w,*tup)

    def setup_widget_from_pango (self, widg, markupstring):
        """setup widget from a pango markup string"""
        #font = pango.FontDescription(fontstring)
        a,t,s = pango.parse_markup(markupstring,u'\x00')
        ai=a.get_iterator()
        font,lang,attrs=ai.get_font()
        return self.setup_widget(widg,font,attrs)

    def setup_widget (self, widg, font, attr):
        tags=self.get_tags_from_attrs(font,None,attr)
        self.tag_widgets[tuple(tags)]=widg
        return widg.connect('toggled',self._toggle,tags)

    def _toggle (self, widget, tags):
        if self.internal_toggle: return
        if widget.get_active():
            for t in tags: self.apply_tag_to_selection(t)
        else:
            for t in tags: self.remove_tag_from_selection(t)

    def _mark_set_cb (self, buffer, iter, mark, *params):
        # Every time the cursor moves, update our widgets that reflect
        # the state of the text.
        if hasattr(self,'_in_mark_set') and self._in_mark_set: return
        self._in_mark_set = True
        if mark.get_name()=='insert':
            for tags,widg in self.tag_widgets.items():
                active = True
                for t in tags:
                    if not iter.has_tag(t):
                        active=False
                self.internal_toggle=True
                widg.set_active(active)
                self.internal_toggle=False
        if hasattr(self,'last_mark'):
            self.move_mark(self.last_mark,iter)
        else:
            self.last_mark = self.create_mark('last',iter,left_gravity=True)
        self._in_mark_set = False

    def _changed_cb (self, tb):
        if not hasattr(self,'last_mark'): return
        # If our insertion point has a mark, we want to apply the tag
        # each time the user types...
        old_itr = self.get_iter_at_mark(self.last_mark)
        insert_itr = self.get_iter_at_mark(self.insert)
        if old_itr!=insert_itr:
            # Use the state of our widgets to determine what
            # properties to apply...
            for tags,w in self.tag_widgets.items():
                if w.get_active():
                    #print 'apply tags...',tags
                    for t in tags: self.apply_tag(t,old_itr,insert_itr)





class SimpleEditor:
    def __init__ (self):
        self.w = gtk.Window()
        self.vb = gtk.VBox()
        self.editBox = gtk.HButtonBox()
        self.nb = gtk.Button('Normal')
        self.editBox.add(self.nb)
        self.sw = gtk.ScrolledWindow()
        self.tv = gtk.TextView()
        self.sw.add(self.tv)
        self.ipb = InteractivePangoBuffer(
            normal_button=self.nb)
        self.ipb.set_text("""<b>This is bold</b>. <i>This is italic</i>
            <b><i>This is bold, italic, and <u>underlined!</u></i></b>
            <span background="blue">This is a test of bg color</span>
            <span foreground="blue">This is a test of fg color</span>
            <span foreground="white" background="blue">This is a test of fg and bg color</span>
            """)
        #    Here are some more: 1-2, 2-3, 3-4, 10-20, 30-40, 50-60
        #    This is <span color="blue">blue</span>, <span color="red">red</span> and <span color="green">green</span>""")
        #self.ipb.set_text("""This is a numerical range (three hundred and fifty to four hundred) 350-400 which may get messed up.
        #Here are some more: 1-2, 2-3, 3-4, 10-20, 30-40, 50-60""")

        self.tv.set_buffer(self.ipb)
        for lab,stock,font in [('gtk-italic',True,'<i>italic</i>'),
                               ('gtk-bold',True,'<b>bold</b>'),
                               ('gtk-underline',True,'<u>underline</u>'),
                               ('Blue',True,'<span foreground="blue">blue</span>'),
                               ('Red',False,'<span foreground="red">smallcaps</span>'),
                               ]:
            button = gtk.ToggleButton(lab)
            self.editBox.add(button)
            if stock: button.set_use_stock(True)
            self.ipb.setup_widget_from_pango(button,font)
        self.vb.add(self.editBox)
        self.vb.add(self.sw)
        self.actionBox = gtk.HButtonBox()
        self.qb = gtk.Button(stock='quit')
        self.pmbut = gtk.Button('Print markup')
        self.pmbut.connect('clicked',self.print_markup)
        self.qb.connect('clicked',lambda *args: self.w.destroy() or gtk.main_quit())
        self.actionBox.add(self.pmbut)
        self.actionBox.add(self.qb)
        self.vb.add(self.actionBox)
        self.w.add(self.vb)
        self.w.show_all()

    def print_markup (self,*args):
        debug.dprint(self.ipb.get_text(), 0)



#!/usr/bin/env python

#    This file is part of Diamond.
#
#    Diamond is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Diamond is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Diamond.  If not, see <http://www.gnu.org/licenses/>.

import gobject
import gtk

class CommentWidget(gtk.Frame):

  __gsignals__ = { "on-store"  : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, ())}

  def __init__(self):
    gtk.Frame.__init__(self)
    
    scrolledWindow = gtk.ScrolledWindow()
    scrolledWindow.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)

    textView = self.textView = gtk.TextView()
    textView.set_editable(False)
    textView.set_wrap_mode(gtk.WRAP_WORD)
    textView.set_cursor_visible(False)
    textView.connect("focus-in-event", self.focus_in)
    textView.connect("focus-out-event", self.focus_out)
    textView.get_buffer().create_tag("tag")
    
    scrolledWindow.add(textView)

    label = gtk.Label()
    label.set_markup("<b>Comment</b>")

    self.set_shadow_type(gtk.SHADOW_NONE)
    self.set_label_widget(label)
    self.add(scrolledWindow)
    
    self.comment_tree = None                           
    return

  def update(self, node):
    """
    Update the widget with the given node
    """

    #before updateing store the old
    self.store()

    if node is None or not node.active:
      self.textView.get_buffer().set_text("")
      self.textView.set_cursor_visible(False)
      self.textView.set_editable(False)
      try:
        self.textView.set_tooltip_text("")
        self.textView.set_property("has-tooltip", False)
      except:
        pass

      return

    self.comment_tree = comment_tree = node.get_comment()
    text_tag = self.textView.get_buffer().get_tag_table().lookup("tag")
    if comment_tree is None:
      self.textView.get_buffer().set_text("No comment")
      self.textView.set_cursor_visible(False)
      self.textView.set_editable(False)
      text_tag.set_property("foreground", "grey")
      try:
        self.textView.set_tooltip_text("")
        self.textView.set_property("has-tooltip", False)
      except:
        pass
    else:
      if comment_tree.data is None:
        self.textView.get_buffer().set_text("(string)")
      else:
        self.textView.get_buffer().set_text(comment_tree.data)
      if node.active:
        self.textView.set_cursor_visible(True)
        self.textView.set_editable(True)
        text_tag.set_property("foreground", "black")
      else:
        self.textView.set_cursor_visible(False)
        self.textView.set_editable(False)
        text_tag.set_property("foreground", "grey")

    buffer_bounds = self.textView.get_buffer().get_bounds()
    self.textView.get_buffer().apply_tag(text_tag, buffer_bounds[0], buffer_bounds[1])

    self.interacted = False

    return

  def store(self):
    """
    Store data in the node comment.
    """
   
    comment_tree = self.comment_tree
    if comment_tree is None or not self.interacted:
      return

    data_buffer_bounds = self.textView.get_buffer().get_bounds()
    new_comment = self.textView.get_buffer().get_text(data_buffer_bounds[0], data_buffer_bounds[1])

    if new_comment != comment_tree.data:
      if new_comment == "":
        comment_tree.data = None
        comment_tree.active = False
      else:
        comment_tree.set_data(new_comment)
        comment_tree.active = True
        self.emit("on-store")
    return

  def focus_in(self, widget, event):
    """
    Called when the comment widget gains focus. Removes the printable_type
    placeholder.
    """

    comment_tree = self.comment_tree
    if not comment_tree is None and not self.interacted:
      self.interacted = True
      if comment_tree.data is None:
        self.textView.get_buffer().set_text("")

    return

  def focus_out(self, widget, event):
    """"
    Called when the comment widget loses focus. Stores the comment.
    """
    self.store()
  
gobject.type_register(CommentWidget)



#!/usr/bin/env python

#    This file is part of Diamond.
#
#    Diamond is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Diamond is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Diamond.  If not, see <http://www.gnu.org/licenses/>.

import os
import os.path
import re
import time
import sys
import tempfile
import cStringIO as StringIO

import pango
import gobject
import gtk
import gtk.glade

import choice
import config
import datatype
import debug
import dialogs
import mixedtree
import plist
import plugins
import schema
import scherror
import tree

import StringIO
import TextBufferMarkup

import attributewidget
import commentwidget
import descriptionwidget
import databuttonswidget
import datawidget
import diffview
import sliceview
import useview

from lxml import etree

try:
  gtk.Tooltip()
except:
  debug.deprint("Interface warning: Unable to use GTK tooltips")

"""
Here are some notes about the code:

Important fields:
  file_path: the directory containing the current file (the working directory or directory of last opened / saved file if no file is open)
  filename: output filename of the current file
  data_paths: paths (from the root node) to important Diamond data (e.g. geometry dimension)
  geometry_dim_tree: MixedTree, with parent equal to the geometry dimension tree and child equal to the geometry dimension data subtree
  gladefile: input Glade file
  gui: GUI GladeXML object
  logofile: the GUI logo file
  main_window: GUI toplevel window
  node_attrs: RHS attributes entry widget
  description: RHS description widget
  data = RHS data widget
  comment: RHS comment entry widget
  node_data: RHS data entry widget
  node_data_buttons_hbox: container for "Revert Data" and "Store Data" buttons
  node_data_interacted: used to determine if a node data widget has been interacted with without data being stored
  node_data_frame: frame containing data entry widgets
  options_tree_select_func_enabled: boolean, true if the options tree select function is enabled (used to overcome a nasty clash with the treeview clicked signal) - re-enabled on next options_tree_select_func call
  selected_node: a tree.Tree or MixedTree containing data to be displayed on the RHS
  selected_iter: last iter set by on_select_row
  s: current schema
  saved: boolean, false if the current file has been edited
  schemafile: the current RNG schema file
  schemafile_path: the directory containing the current schema file (the working directory if no schema is open)
  signals: dictionary containing signal handlers for signals set up in the Glade file
  statusbar: GUI status bar
  tree: LHS tree root
  treestore: the LHS tree model
  treeview: the LHS tree widget

Important routines:
  cellcombo_changed: called when a choice is selected on the left-hand pane
  init_treemodel: set up the treemodel and treeview
  on_treeview_clicked: when a row is clicked, process the consequences (e.g. activate inactive instance)
  set_treestore: stuff the treestore with a given tree.Tree
  on_find_find_button & search_treestore: the find functionality
  on_select_row: when a row is selected, update the options frame
  update_options_frame: paint the right-hand side

If there are bugs in reading in, see schema.read.
If there are bugs in writing out, see tree.write.
"""

class Diamond:
  def __init__(self, gladefile, schemafile = None, schematron_file = None, logofile = None, input_filename = None, 
      dim_path = "/geometry/dimension", suffix=None):
    self.gladefile = gladefile
    self.gui = gtk.glade.XML(self.gladefile)

    self.statusbar = DiamondStatusBar(self.gui.get_widget("statusBar"))
    self.find      = DiamondFindDialog(self, gladefile)
    self.popup = self.gui.get_widget("popupmenu")

    self.add_custom_widgets()
    
    self.plugin_buttonbox = self.gui.get_widget("plugin_buttonbox")
    self.plugin_buttonbox.set_layout(gtk.BUTTONBOX_START)
    self.plugin_buttonbox.show()
    self.plugin_buttons = []

    self.scherror  = scherror.DiamondSchemaError(self, gladefile, schemafile, schematron_file)

    signals     =  {"on_new": self.on_new,
                    "on_quit": self.on_quit,
                    "on_open": self.on_open,
                    "on_open_schema": self.on_open_schema,
                    "on_save": self.on_save,
                    "on_save_as": self.on_save_as,
                    "on_validate": self.scherror.on_validate,
                    "on_validate_schematron": self.scherror.on_validate_schematron,
                    "on_expand_all": self.on_expand_all,
                    "on_collapse_all": self.on_collapse_all,
                    "on_find": self.find.on_find,
                    "on_go_to_node": self.on_go_to_node,
                    "on_console": self.on_console,
                    "on_display_properties_toggled": self.on_display_properties_toggled,
                    "on_about": self.on_about,
                    "on_copy_spud_path": self.on_copy_spud_path,
                    "on_copy": self.on_copy,
                    "on_paste": self.on_paste,
                    "on_slice": self.on_slice,
                    "on_diff": self.on_diff,
                    "on_diffsave": self.on_diffsave,
                    "on_finduseage": self.on_finduseage,
                    "on_group": self.on_group,
                    "on_ungroup": self.on_ungroup}

    self.gui.signal_autoconnect(signals)

    self.main_window = self.gui.get_widget("mainWindow")
    self.main_window.connect("delete_event", self.on_delete)

    self.logofile = logofile
    if self.logofile is not None:
      for l in self.logofile:
		  try:
			gtk.window_set_default_icon_from_file(l)
			break
		  except:
		    pass

    self.init_treemodel()

    self.data_paths = {}
    self.data_paths["dim"] = dim_path

    self.suffix = suffix

    self.selected_node = None
    self.selected_iter = None
    self.update_options_frame()

    self.file_path = os.getcwd()
    self.schemafile_path = os.getcwd()
    self.filename = None
    self.schemafile = None
    self.init_datatree()    
    self.set_saved(True)
    self.open_file(schemafile = None, filename = None)

    self.main_window.show()

    if schemafile is not None:
      self.open_file(schemafile = schemafile, filename = input_filename)

    # Hide specific menu items
    menu = self.gui.get_widget("menu")
    
    # Disable Find
    menu.get_children()[1].get_submenu().get_children()[0].set_property("visible", False)

    if schematron_file is None:
      # Disable Validate Schematron
      menu.get_children()[3].get_submenu().get_children()[1].set_property("sensitive", False)

    return

  def program_exists(self, name):
    ret = os.system("which %s > /dev/null" % name)
    return ret == 0

  ### MENU ###

  def update_title(self):
    """
    Update the Diamond title based on the save status of the currently open file.
    """

    title = "Diamond: "
    if not self.saved:
      title += "*"
    if self.filename is None:
      title += "(Unsaved)"
    else:
      title += os.path.basename(self.filename)
      if len(os.path.dirname(self.filename)) > 0:
        title += " (%s)" % os.path.dirname(self.filename)

    self.main_window.set_title(title)

    return

  def set_saved(self, saved, filename = ""):
    """
    Change the save status of the current file.
    """

    self.saved = saved
    if filename != "":
      self.filename = filename
      if filename is not None:
        self.file_path = os.path.dirname(filename) + os.path.sep
    self.update_title()

    return

  def close_schema(self):
    if self.schemafile is None:
      return

    # clear the schema.
    self.s = None
    self.schemafile = None
    self.schemafile_path = None
    self.scherror.schema_file = None
       
    return

  def load_schema(self, schemafile):
    # so, if the schemafile has already been opened, then ..
    if schemafile == self.schemafile:
      self.statusbar.set_statusbar('Schema ' + schemafile + ' already loaded')
      return

    # if we aren't using a http schema, and we're passed a relative filename, we
    # need to absolut-ify it.
    if 'http' not in schemafile:
      schemafile = os.path.abspath(schemafile)

    self.statusbar.set_statusbar('Loading schema from ' + schemafile)

    # now, let's try and read the schema.
    try:
      s_read = schema.Schema(schemafile)
      self.s = s_read
      self.statusbar.set_statusbar('Loaded schema from ' + schemafile)
    except:
      dialogs.error_tb(self.main_window, "Unable to open schema file \"" + schemafile + "\"")
      self.statusbar.clear_statusbar()
      return

    self.schemafile = schemafile
    self.schemafile_path = os.path.dirname(schemafile) + os.path.sep
    self.scherror.schema_file = schemafile
    
    self.remove_children(None)
    self.init_datatree()
    
    return

  def close_file(self):
    self.remove_children(None)
    self.init_datatree()

    self.filename = None
    
    return

  def display_validation_errors(self, errors):
    # Extract and display validation errors
    saved = True
    lost_eles, added_eles, lost_attrs, added_attrs = errors
    if len(lost_eles) > 0 or len(added_eles) > 0 or len(lost_attrs) > 0 or len(added_attrs) > 0:
      saved = False
      msg = ""
      if len(lost_eles) > 0:
        msg += "Warning: lost xml elements:\n"
        for ele in lost_eles:
          msg += ele + "\n"
      if len(added_eles) > 0:
        msg += "Warning: added xml elements:\n"
        for ele in added_eles:
          msg += ele + "\n"
      if len(lost_attrs) > 0:
        msg += "Warning: lost xml attributes:\n"
        for ele in lost_attrs:
          msg += ele + "\n"
      if len(added_attrs) > 0:
        msg += "Warning: added xml attributes:\n"
        for ele in added_attrs:
          msg += ele + "\n"

      dialogs.long_message(self.main_window, msg)
    return saved

  def load_file(self, filename):
    # if we have a relative path, make it absolute
    filename = os.path.abspath(filename)

    try:
      os.stat(filename)
    except OSError:
      self.filename = filename
      self.set_saved(False)

      self.remove_children(None)
      self.init_datatree()

      return

    try:
      tree_read = self.s.read(filename)

      if tree_read is None:
        dialogs.error_tb(self.main_window, "Unable to open file \"" + filename + "\"")
        return

      saved = self.display_validation_errors(self.s.read_errors())

      self.tree = tree_read
      self.filename = filename
    except:
      dialogs.error_tb(self.main_window, "Unable to open file \"" + filename + "\"")
      return

    self.set_saved(saved, filename)

    return

  def open_file(self, schemafile = "", filename = ""):
    """
    Handle opening or clearing of the current file and / or schema.
    """

    self.find.on_find_close_button()
    if schemafile is None:
      self.close_schema()
    elif schemafile != "":
      self.load_schema(schemafile)
    if filename is None:
      self.close_file()
    elif filename != "":
      self.load_file(filename)
      
    self.treeview.freeze_child_notify()
    self.treeview.set_model(None)
    self.signals = {}
    self.set_treestore(None, [self.tree], True)
    self.treeview.set_model(self.treestore)
    self.treeview.thaw_child_notify()

    self.set_geometry_dim_tree()

    self.treeview.grab_focus()
    self.treeview.get_selection().select_path(0)

    self.selected_node = None
    self.update_options_frame()

    self.scherror.destroy_error_list()

    return

  def save_continue(self):

    if not self.saved:
      prompt_response = dialogs.prompt(self.main_window, 
        "Unsaved data. Do you want to save the current document before continuing?", gtk.MESSAGE_WARNING, True)
 
      if prompt_response == gtk.RESPONSE_YES:
        if self.filename is None:
          return self.on_save_as()
        else:
          return self.on_save()
      elif prompt_response == gtk.RESPONSE_CANCEL:
        return False

    return True

  def on_new(self, widget=None):
    """
    Called when new is clicked. Clear the treestore and reset the datatree.
    """

    if not self.save_continue():
      return

    self.open_file(filename = None)
    self.filename = None

    return

  def on_open(self, widget=None):
    """
    Called when open is clicked. Open a user supplied file.
    """

    if not self.save_continue():
      return

    filter_names_and_patterns = {}
    if self.suffix is None:
      for xmlname in config.schemata:
        filter_names_and_patterns[config.schemata[xmlname][0]] = "*." + xmlname
    elif self.suffix in config.schemata.keys():
      filter_names_and_patterns[config.schemata[self.suffix][0]] = "*." + self.suffix
    else:
      filter_names_and_patterns[self.suffix] = "*." + self.suffix

    filename = dialogs.get_filename(title = "Open XML file", action = gtk.FILE_CHOOSER_ACTION_OPEN, filter_names_and_patterns = filter_names_and_patterns, folder_uri = self.file_path)

    if filename is not None:
      self.open_file(filename = filename)

    return

  def on_open_schema(self, widget=None):
    """
    Called when open schema is clicked. Clear the treestore and reset the schema.
    """

    if not self.save_continue():
      return

    filename = dialogs.get_filename(title = "Open RELAX NG schema", action = gtk.FILE_CHOOSER_ACTION_OPEN, filter_names_and_patterns = {"RNG files":"*.rng"}, folder_uri = self.schemafile_path)
    if filename is not None:
      self.open_file(schemafile = filename)

    return

  def on_save(self, widget=None):
    """
    Write out to XML. If we don't already have a filename, open a dialog to get
    one.
    """

    self.data.store()

    if self.filename is None:
      return self.on_save_as(widget)
    else:
      self.statusbar.set_statusbar("Saving ...")
      self.main_window.window.set_cursor(gtk.gdk.Cursor(gtk.gdk.WATCH))
      try:
        self.tree.write(self.filename)
      except:
        dialogs.error_tb(self.main_window, "Saving to \"" + self.filename + "\" failed")
        self.statusbar.clear_statusbar()
        self.main_window.window.set_cursor(None)
        return False

      self.set_saved(True)

      self.statusbar.clear_statusbar()
      self.main_window.window.set_cursor(None)
      return True

    return False

  def on_save_as(self, widget=None):
    """
    Write out the XML to a file.
    """

    if self.schemafile is None:
      dialogs.error(self.main_window, "No schema file open")
      return False

    filter_names_and_patterns = {}
    if self.suffix is None:
      for xmlname in config.schemata:
        filter_names_and_patterns[config.schemata[xmlname][0]] = "*." + xmlname
    elif self.suffix in config.schemata.keys():
      filter_names_and_patterns[config.schemata[self.suffix][0]] = "*." + self.suffix
    else:
      filter_names_and_patterns[self.suffix] = "*." + self.suffix

    filename = dialogs.get_filename(title = "Save XML file", action = gtk.FILE_CHOOSER_ACTION_SAVE, filter_names_and_patterns = filter_names_and_patterns, folder_uri = self.file_path)

    if filename is not None:
      # Check that the selected file has a file extension. If not, add a .xml extension.
      if len(filename.split(".")) <= 1:
        filename += ".xml"

      # Save the file
      self.statusbar.set_statusbar("Saving ...")
      self.main_window.window.set_cursor(gtk.gdk.Cursor(gtk.gdk.WATCH))
      self.tree.write(filename)
      self.set_saved(True, filename)
      self.statusbar.clear_statusbar()
      self.main_window.window.set_cursor(None)
      return True

    return False

  def on_delete(self, widget, event):
    """
    Called when the main window is deleted. Return "True" to prevent the deletion
    of the main window (deletion is handled by "on_quit").
    """

    self.on_quit(widget, event)

    return True

  def on_quit(self, widget, event = None):
    """
    Quit the program. Prompt the user to save data if the current file has been
    changed.
    """

    if not self.save_continue():
      return

    self.destroy()

    return

  def destroy(self):
    """
    End the program.
    """

    try:
      gtk.main_quit()
    except:
      debug.dprint("Failed to quit - already quit?")

    return

  def on_display_properties_toggled(self, widget=None):
    optionsFrame = self.gui.get_widget("optionsFrame")
    optionsFrame.set_property("visible", not optionsFrame.get_property("visible"))
    return

  def on_go_to_node(self, widget=None):
   """
   Go to a node, identified by an XPath
   """

   dialog = dialogs.GoToDialog(self)
   spudpath = dialog.run()

   return

  def on_expand_all(self, widget=None):
    """
    Show the whole tree.
    """

    self.treeview.expand_all()

    return

  def on_collapse_all(self, widget=None):
    """
    Collapse the whole tree.
    """

    self.treeview.collapse_all()

    return
    
  def on_console(self, widget = None):
    """
    Launch a python console
    """    
    
    # Construct the dictionary of locals that will be used by the interpreter
    locals = {}
    locals["interface"] = globals()
    locals["diamond_gui"] = self
  
    dialogs.console(self.main_window, locals)
    
    return

  def on_about(self, widget=None):
    """
    Tell the user how fecking great we are.
    """

    about = gtk.AboutDialog()
    about.set_name("Diamond")
    about.set_copyright("GPLv3")
    about.set_comments("A RELAX-NG-aware XML editor")
    about.set_authors(["Patrick E. Farrell", "James R. Maddison", "Matthew T. Whitworth", "Fraser J. Waters"])
    about.set_license("Diamond is free software: you can redistribute it and/or modify\n"+
                      "it under the terms of the GNU General Public License as published by\n"+
                      "the Free Software Foundation, either version 3 of the License, or\n"+
                      "(at your option) any later version.\n"+
                      "\n"+
                      "Diamond is distributed in the hope that it will be useful,\n"+
                      "but WITHOUT ANY WARRANTY; without even the implied warranty of\n"+
                      "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"+
                      "GNU General Public License for more details.\n"+
                      "You should have received a copy of the GNU General Public License\n"+
                      "along with Diamond.  If not, see http://www.gnu.org/licenses/.")

    if self.logofile is not None:
      for l in self.logofile:
        try:
          logo = gtk.gdk.pixbuf_new_from_file(l)
          about.set_logo(logo)
          break
        except:
          pass
        
      
    try:
      image = about.get_children()[0].get_children()[0].get_children()[0]
      image.set_tooltip_text("Diamond: it's clearer than GEM")
    except:
      pass
    
    about.run()
    about.destroy()

    return


  def on_copy_spud_path(self, widget=None):
    path = self.get_selected_row(self.treeview.get_selection())
    if path is None:
      debug.deprint("No selection.")
      return
    iter = self.treestore.get_iter(path)
    active_tree = self.treestore.get_value(iter, 1)
    name = self.get_spudpath(active_tree)
    clipboard = gtk.clipboard_get()
    clipboard.set_text(name)
    clipboard.store()

  def _get_focus_widget(self, parent):
    """
    Gets the widget that is a child of parent with the focus.
    """
    focus = parent.get_focus_child()
    if focus is None or (focus.flags() & gtk.HAS_FOCUS):
      return focus
    else:
      return self._get_focus_widget(focus)

  def _handle_clipboard(self, widget, signal):
    """
    This finds the currently focused widget.
    If no widget is focused or the focused widget doesn't support 
    the given clipboard operation use the treeview (False), otherwise
    signal the widget to handel the clipboard operation (True).
    """
    widget = self._get_focus_widget(self.main_window)

    if widget is None or widget is self.treeview:
      return False

    if gobject.signal_lookup(signal + "-clipboard", widget):
      widget.emit(signal + "-clipboard")
      return True
    else:
      return False

  def on_copy(self, widget=None):
    if self._handle_clipboard(widget, "copy"):
      return

    if isinstance(self.selected_node, mixedtree.MixedTree):
      node = self.selected_node.parent
    else:
      node = self.selected_node    

    if node != None and node.active:
      ios = StringIO.StringIO()
      node.write(ios)
    
      clipboard = gtk.clipboard_get()
      clipboard.set_text(ios.getvalue())
      clipboard.store()

      ios.close()
    return

  def on_paste(self, widget=None):
    if self._handle_clipboard(widget, "paste"):
      return

    clipboard = gtk.clipboard_get()
    ios = StringIO.StringIO(clipboard.wait_for_text())
    
    if self.selected_iter is not None:    
      node = self.treestore.get_value(self.selected_iter, 0)

    if node != None:
  
      expand = not node.active
      if expand:
        self.expand_tree(self.selected_iter)

      newnode = self.s.read(ios, node)

      if newnode is None:
        if expand:
          self.collapse_tree(self.selected_iter, False)
        self.statusbar.set_statusbar("Trying to paste invalid XML.")
        return

      if node.parent is not None:
        newnode.set_parent(node.parent)
        children = node.parent.get_children()
        children.insert(children.index(node), newnode)
        children.remove(node)
 
      self.treeview.freeze_child_notify()
      iter = self.set_treestore(self.selected_iter, [newnode], True, True)
      newnode.recompute_validity()
      self.treeview.thaw_child_notify()

      self.treeview.get_selection().select_iter(iter)

      self.display_validation_errors(self.s.read_errors())

      self.set_saved(False)

    return

  def __diff(self, path):
    def run_diff(self, path):
      start = time.clock()
      diffview.DiffView(path, self.tree)
      seconds = time.clock() - start
      self.statusbar.set_statusbar("Diff calculated (took " + str(seconds) + " seconds)")
      return False

    if path and os.path.isfile(path):
      filename = path
    else:
      dialog = gtk.FileChooserDialog(
                                     title = "Diff against",
                                     action = gtk.FILE_CHOOSER_ACTION_OPEN,
                                     buttons = (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL, gtk.STOCK_OPEN, gtk.RESPONSE_OK))

      if path:
        dialog.set_current_folder(path)

      response = dialog.run()
      if response != gtk.RESPONSE_OK:
        dialog.destroy()
        return

      filename = dialog.get_filename()
      dialog.destroy()

    self.statusbar.set_statusbar("Calculating diff... (this may take a while)")
    gobject.idle_add(run_diff, self, filename)

  def on_diff(self, widget = None, path = None):
    if path is None:
      path = os.path.dirname(self.filename) if self.filename else None
    self.__diff(path)

  def on_diffsave(self, widget = None):
    if self.filename:
      self.__diff(self.filename)
    else:
      dialogs.error(self.main_window, "No save to diff against.")

  def on_finduseage(self, widget = None):
    useview.UseView(self.s, self.suffix)

  def on_slice(self, widget = None):
    if not self.selected_node.is_sliceable():
      self.statusbar.set_statusbar("Cannot slice on this element.")
      return

    window = sliceview.SliceView(self.main_window)
    window.geometry_dim_tree = self.geometry_dim_tree
    window.update(self.selected_node, self.tree)
    window.connect("destroy", self._slice_destroy)
    return

  def _slice_destroy(self, widget):
    self.on_select_row()


  groupmode = False

  def on_group(self, widget = None):
    """
    Clears the treeview and then fills it with nodes 
    with the same type as the selected node.
    """

    if self.selected_node == self.tree or not self.selected_iter:
      self.statusbar.set_statusbar("Cannot group on this element.")
      return #Group on the entire tree... ie don't group or nothing selected

    self.gui.get_widget("menuitemUngroup").show()
    self.gui.get_widget("popupmenuitemUngroup").show()

    self.groupmode = True
    node, tree = self.treestore.get(self.selected_iter, 0, 1)
    
    self.treeview.freeze_child_notify()
    self.treeview.set_model(None)

    def get_nodes(node, tree):
      nodes = []

      if isinstance(tree, choice.Choice):
        child = tree.get_current_tree()
        if child.name == node.name:
          nodes.append(tree)
        nodes += get_nodes(node, child)
      else:
        for child in tree.get_children():
          if child.name == node.name:
            nodes.append(child)
          nodes += get_nodes(node, child)

      return nodes
    
    self.set_treestore(None, get_nodes(tree, self.tree), True)

    self.treeview.set_model(self.treestore)
    self.treeview.thaw_child_notify()

    path = self.get_treestore_path_from_node(node)
    self.treeview.get_selection().select_path(path)

    return

  def on_ungroup(self, widget = None):
    """
    Restores the treeview to normal.
    """

    self.gui.get_widget("menuitemUngroup").hide()
    self.gui.get_widget("popupmenuitemUngroup").hide()

    self.groupmode = False
    node = self.treestore.get_value(self.selected_iter, 0)

    self.treeview.freeze_child_notify()
    self.treeview.set_model(None)

    self.set_treestore(None, [self.tree], True)

    self.treeview.set_model(self.treestore)
    self.treeview.thaw_child_notify()

    path = self.get_treestore_path_from_node(node)
    self.treeview.expand_to_path(path)
    self.treeview.scroll_to_cell(path)
    self.treeview.get_selection().select_path(path)
 
    return

  ## LHS ###

  def init_datatree(self):
    """
    Add the root node of the XML tree to the treestore, and its children.
    """

    if self.schemafile is None:
      self.set_treestore(None, [])
      self.tree = None
    else:
      l = self.s.valid_children(":start")

      self.tree = l[0]
      self.signals = {}
      self.set_treestore(None, l)

    root_iter = self.treestore.get_iter_first()
    self.treeview.freeze_child_notify()
    self.treeview.set_model(None)
    self.expand_treestore(root_iter)
    self.treeview.set_model(self.treestore)
    self.treeview.thaw_child_notify()

    return

  def init_treemodel(self):
    """
    Set up the treestore and treeview.
    """

    self.treeview = optionsTree = self.gui.get_widget("optionsTree")
    self.treeview.connect("key_press_event", self.on_treeview_key_press)
    self.treeview.connect("button_press_event", self.on_treeview_button_press)
    self.treeview.connect("row-activated", self.on_activate_row)
    self.treeview.connect("popup_menu", self.on_treeview_popup)

    self.treeview.set_property("rules-hint", True)

    self.treeview.get_selection().set_mode(gtk.SELECTION_SINGLE)
    self.treeview.get_selection().connect("changed", self.on_select_row)
    self.treeview.get_selection().set_select_function(self.options_tree_select_func)
    self.options_tree_select_func_enabled = True

    # Node column
    column = gtk.TreeViewColumn("Node")
    column.set_property("expand", True)
    column.set_resizable(True)

    self.cellcombo = cellCombo = gtk.CellRendererCombo()
    cellCombo.set_property("text-column", 0)
    cellCombo.set_property("editable", True)
    cellCombo.set_property("has-entry", False)
    cellCombo.connect("changed", self.cellcombo_changed)
    column.pack_start(cellCombo)
    column.set_cell_data_func(cellCombo, self.set_combobox_liststore)

    self.choicecell = choiceCell = gtk.CellRendererPixbuf()
    column.pack_end(choiceCell, expand=False)
    column.set_cell_data_func(choiceCell, self.set_cellpicture_choice)

    optionsTree.append_column(column)

    self.imgcell = cellPicture = gtk.CellRendererPixbuf()
    self.imgcolumn = imgcolumn = gtk.TreeViewColumn("", cellPicture)
    imgcolumn.set_property("expand", False)
    imgcolumn.set_property("fixed-width", 20)
    imgcolumn.set_property("sizing", gtk.TREE_VIEW_COLUMN_FIXED)
    imgcolumn.set_cell_data_func(cellPicture, self.set_cellpicture_cardinality)
    optionsTree.append_column(imgcolumn)

    # 0: pointer to node in self.tree -- a choice or a tree
    # 1: pointer to currently active tree
    # 2: gtk.ListStore containing the display names of possible choices

    self.treestore = gtk.TreeStore(gobject.TYPE_PYOBJECT, gobject.TYPE_PYOBJECT, gobject.TYPE_PYOBJECT)
    self.treeview.set_model(self.treestore)
    self.treeview.set_enable_search(False)

    return

  def create_liststore(self, choice_or_tree):
    """
    Given a list of possible choices, create the liststore for the
    gtk.CellRendererCombo that contains the names of possible choices.
    """

    liststore = gtk.ListStore(str, gobject.TYPE_PYOBJECT)

    for t in choice_or_tree.get_choices():
      liststore.append([str(t), t])

    return liststore
  
  def set_treestore(self, iter=None, new_tree=[], recurse=False, replace=False):
    """
    Given a list of children of a node in a treestore, stuff them in the treestore.
    """

    if replace:
      replacediter = iter
      iter = self.treestore.iter_parent(replacediter)
    else: 
      self.remove_children(iter)
  
    for t in new_tree:
      if t is not None and t in self.signals:
        t.disconnect(self.signals[t][0])
        t.disconnect(self.signals[t][1])

      if isinstance(t, tree.Tree):
        if t.is_hidden():
          attrid = t.connect("on-set-attr", self.on_set_attr, self.treestore.get_path(iter))
          dataid = t.connect("on-set-data", self.on_set_data, self.treestore.get_path(iter))
          self.signals[t] = (attrid, dataid)
          continue

        liststore = self.create_liststore(t)

        if replace:
          child_iter = self.treestore.insert_before(iter, replacediter, [t, t, liststore])
        else:
          child_iter = self.treestore.append(iter, [t, t, liststore])

        attrid = t.connect("on-set-attr", self.on_set_attr, self.treestore.get_path(child_iter))
        dataid = t.connect("on-set-data", self.on_set_data, self.treestore.get_path(child_iter))
        self.signals[t] = (attrid, dataid)
 
        if recurse and t.active: self.set_treestore(child_iter, t.children, recurse)

      elif isinstance(t, choice.Choice):
        liststore = self.create_liststore(t)
        ts_choice = t.get_current_tree()
        if ts_choice.is_hidden():
          attrid = t.connect("on-set-attr", self.on_set_attr, self.treestore.get_path(iter))
          dataid = t.connect("on-set-data", self.on_set_data, self.treestore.get_path(iter))
          self.signals[t] = (attrid, dataid)
          continue

        if replace:
          child_iter = self.treestore.insert_before(iter, replacediter, [t, ts_choice, liststore])
        else:
          child_iter = self.treestore.append(iter, [t, ts_choice, liststore])

        attrid = t.connect("on-set-attr", self.on_set_attr, self.treestore.get_path(child_iter))
        dataid = t.connect("on-set-data", self.on_set_data, self.treestore.get_path(child_iter))
        self.signals[t] = (attrid, dataid)

        if recurse and t.active: self.set_treestore(child_iter, ts_choice.children, recurse)

    if replace:
      self.treestore.remove(replacediter)
      return child_iter
   
    return iter

  def expand_choice_or_tree(self, choice_or_tree):
    """
    Query the schema for what valid children can live under this node, and add
    them to the choice or tree. This recurses.
    """

    if isinstance(choice_or_tree, choice.Choice):
      for opt in choice_or_tree.choices():
        self.expand_choice_or_tree(opt)
    else:
      l = self.s.valid_children(choice_or_tree.schemaname)
      l = choice_or_tree.find_or_add(l)
      for opt in l:
        self.expand_choice_or_tree(opt)

    return

  def expand_treestore(self, iter = None):
    """
    Query the schema for what valid children can live under this node, then set the
    treestore appropriately. This recurses.
    """

    if iter is None:
      iter = self.treestore.get_iter_first()
      if iter is None:
        self.set_treestore(iter, [])
        return

    choice_or_tree, active_tree = self.treestore.get(iter, 0, 1)
    if active_tree.active is False or choice_or_tree.active is False:
      return

    l = self.s.valid_children(active_tree.schemaname)
    l = active_tree.find_or_add(l)
    self.set_treestore(iter, l)

    child_iter = self.treestore.iter_children(iter)
    while child_iter is not None:
      # fix for recursive schemata!
      child_active_tree = self.treestore.get_value(child_iter, 1)
      if child_active_tree.schemaname == active_tree.schemaname:
        debug.deprint("Warning: recursive schema elements not supported: %s" % active_tree.name)
        child_iter = self.treestore.iter_next(child_iter)
        if child_iter is None: break

      self.expand_treestore(child_iter)
      child_iter = self.treestore.iter_next(child_iter)

    return

  def remove_children(self, iter):
    """
    Delete the children of iter in the treestore.
    """

    childiter = self.treestore.iter_children(iter)
    if childiter is None: return

    result = True

    while result is True:
      result = self.treestore.remove(childiter)

    return

  def set_combobox_liststore(self, column, cellCombo, treemodel, iter, user_data=None):
    """
    This hook function sets the properties of the gtk.CellRendererCombo for each
    row. It sets up the cellcombo to use the correct liststore for its choices,
    decides whether the cellcombo should be editable or not, and sets the
    foreground colour.
    """

    choice_or_tree, active_tree, liststore = treemodel.get(iter, 0, 1, 2)

    if self.groupmode and treemodel.iter_parent(iter) is None:
      cellCombo.set_property("text", choice_or_tree.get_name_path())
    else:
      cellCombo.set_property("text", str(choice_or_tree))


    cellCombo.set_property("model", liststore)

    # set the properties: colour, etc.
    if isinstance(choice_or_tree, tree.Tree):
      cellCombo.set_property("editable", False)
    elif isinstance(choice_or_tree, choice.Choice):
      cellCombo.set_property("editable", True)

    if self.iter_is_active(treemodel, iter):
      if active_tree.valid is True:
        cellCombo.set_property("foreground", "black")
      else:
        cellCombo.set_property("foreground", "blue")
    else:
        cellCombo.set_property("foreground", "gray")

    return

  def set_cellpicture_choice(self, column, cell, treemodel, iter):
    """
    This hook function sets up the other gtk.CellRendererPixbuf, the one that gives
    the clue to the user whether this is a choice or not.
    """
    
    choice_or_tree = treemodel.get_value(iter, 0)
    if isinstance(choice_or_tree, tree.Tree):
      cell.set_property("stock-id", None)
    elif isinstance(choice_or_tree, choice.Choice):
      cell.set_property("stock-id", gtk.STOCK_GO_DOWN)

    return

  def set_cellpicture_cardinality(self, column, cell, treemodel, iter):
    """
    This hook function sets up the gtk.CellRendererPixbuf on the extreme right-hand
    side for each row; this paints a plus or minus or nothing depending on whether
    something can be added or removed or has to be there.
    """

    choice_or_tree = treemodel.get_value(iter, 0)
    if choice_or_tree.cardinality == "":
      cell.set_property("stock-id", None)
    elif choice_or_tree.cardinality == "?" or choice_or_tree.cardinality == "*":
      if choice_or_tree.active:
        cell.set_property("stock-id", gtk.STOCK_REMOVE)
      else:
        cell.set_property("stock-id", gtk.STOCK_ADD)
    elif choice_or_tree.cardinality == "+":
      parent_tree = choice_or_tree.parent
      count = parent_tree.count_children_by_schemaname(choice_or_tree.schemaname)
      
      if choice_or_tree.active and count == 2: # one active, one inactive
        cell.set_property("stock-id", None)
      elif choice_or_tree.active:
        cell.set_property("stock-id", gtk.STOCK_REMOVE)
      else:
        cell.set_property("stock-id", gtk.STOCK_ADD)

    return

  def toggle_tree(self, iter):
    """
    Toggles the state of part of the tree.
    """

    choice_or_tree = self.treestore.get_value(iter, 0)

    if choice_or_tree.active:
      self.collapse_tree(iter)
    else:
      self.expand_tree(iter)

    self.on_select_row()

    return
  
  def collapse_tree(self, iter, confirm = True):
    """
    Collapses part of the tree.
    """

    choice_or_tree, = self.treestore.get(iter, 0)
    parent_tree = choice_or_tree.parent

    if not choice_or_tree.active:
      return

    if choice_or_tree.cardinality == "":
      return

    if choice_or_tree.cardinality == "?":
      choice_or_tree.active = False
      self.set_saved(False)
      self.remove_children(iter)

    elif choice_or_tree.cardinality == "*":
      # If this is the only one, just make it inactive.
      # Otherwise, just delete it.
      count = parent_tree.count_children_by_schemaname(choice_or_tree.schemaname)
      if count == 1:
        choice_or_tree.active = False
        self.set_saved(False)
        self.remove_children(iter)
      else:
        self.delete_tree(iter, confirm)

    elif choice_or_tree.cardinality == "+":
      count = parent_tree.count_children_by_schemaname(choice_or_tree.schemaname)
      if count == 2: # one active, one inactive
        # do nothing
        return
      else: # count > 2
        self.delete_tree(iter, confirm)
  
    parent_tree.recompute_validity()
    self.treeview.queue_draw()
    return

  def delete_tree(self, iter, confirm):
    choice_or_tree, = self.treestore.get(iter, 0)
    parent_tree = choice_or_tree.parent
    isSelected = self.treeview.get_selection().iter_is_selected(iter)
    sibling = self.treestore.iter_next(iter)

    if confirm:
      response = dialogs.prompt(self.main_window, "Are you sure you want to delete this node?")

    # not A or B == A implies B
    if not confirm or response == gtk.RESPONSE_YES:
      parent_tree.delete_child_by_ref(choice_or_tree)
      self.remove_children(iter)
      self.treestore.remove(iter)
      self.set_saved(False)
      
      if isSelected and sibling:
        self.treeview.get_selection().select_iter(sibling)
    return

  def expand_tree(self, iter):
    """
    Expands part of the tree.
    """

    choice_or_tree, active_tree = self.treestore.get(iter, 0, 1)
    parent_tree = choice_or_tree.parent

    if choice_or_tree.active:
      return

    if choice_or_tree.cardinality == "":
      return

    elif choice_or_tree.cardinality == "?":
      choice_or_tree.active = True
      self.set_saved(False)
      self.expand_treestore(iter)

    elif choice_or_tree.cardinality == "*" or choice_or_tree.cardinality == "+":
      # Make this active, and add a new inactive instance
      choice_or_tree.active = True
      new_tree = parent_tree.add_inactive_instance(choice_or_tree)
      liststore = self.create_liststore(new_tree)
      self.expand_treestore(iter)
      iter = self.treestore.insert_after(
        None, iter, [new_tree, new_tree.get_current_tree(), liststore])
      attrid = new_tree.connect("on-set-attr", self.on_set_attr, self.treestore.get_path(iter))
      dataid = new_tree.connect("on-set-data", self.on_set_data, self.treestore.get_path(iter))
      self.signals[new_tree] = (attrid, dataid)
      self.set_saved(False)

    parent_tree.recompute_validity()
    return

  def options_tree_select_func(self, info = None):
    """
    Called when the user selected a new item in the treeview. Prevents changing of
    node and attempts to save data if appropriate.
    """

    if not self.options_tree_select_func_enabled:
      self.options_tree_select_func_enabled = True
      return False

    if not self.data.store():
      return False

    if (isinstance(self.selected_node, mixedtree.MixedTree)
       and self.geometry_dim_tree is not None
       and self.selected_node.parent is self.geometry_dim_tree.parent
       and self.selected_node.data is not None):
      self.geometry_dim_tree.set_data(self.selected_node.data)

    return True

  def on_treeview_key_press(self, treeview, event):
    """
    Called when treeview intercepts a key press. Collapse and expand rows.
    """

    if event.keyval == gtk.keysyms.Right:
      self.treeview.expand_row(self.get_selected_row(), open_all = False)

    if event.keyval == gtk.keysyms.Left:
      self.treeview.collapse_row(self.get_selected_row())

    if event.keyval == gtk.keysyms.Delete:
       self.collapse_tree(self.treestore.get_iter(self.get_selected_row()))
       self.on_select_row()
 
    return

  def on_treeview_button_press(self, treeview, event):
    """
    This routine is called every time the mouse is clicked on the treeview on the
    left-hand side. It processes the "buttons" gtk.STOCK_ADD and gtk.STOCK_REMOVE
    in the right-hand column, activating, adding and removing tree nodes as
    necessary.
    """
    pathinfo = treeview.get_path_at_pos(int(event.x), int(event.y))

    if event.button == 1:

      if pathinfo is not None:
        path = pathinfo[0]
        col = pathinfo[1]

        if col is self.imgcolumn:
          iter = self.treestore.get_iter(path)
          self.toggle_tree(iter)

    elif event.button == 3:
      if pathinfo is not None:
        treeview.get_selection().select_path(pathinfo[0])
        self.show_popup(None, event.button, event.time)
        return True

  def popup_location(self, widget):
    column = self.treeview.get_column(0)
    treemodel, treeiter = self.treeview.get_selection().get_selected()
    treepath = treemodel.get_path(treeiter)
    area = self.treeview.get_cell_area(treepath, column)
    sx, sy = self.popup.size_request()
    tx, ty = area.x, area.y + int(sy * 1.25)
    x, y = self.treeview.tree_to_widget_coords(tx, ty)
    return (x, y, True)

  def on_treeview_popup(self, treeview):
    self.show_popup(self.popup_location, 0, gtk.get_current_event_time())
    return

  def show_popup(self, func, button, time):
    self.popup.popup(None, None, func, button, time)
    return

  def on_select_row(self, selection=None):
    """
    Called when a row is selected. Update the options frame.
    """

    if isinstance(selection, str) or isinstance(selection, tuple):
      path = selection
    else:
      path = self.get_selected_row(selection)

    if path is None:
      return  

    self.selected_iter = iter = self.treestore.get_iter(path)
    choice_or_tree, active_tree = self.treestore.get(iter, 0, 1)

    debug.dprint(active_tree)

    self.selected_node = self.get_painted_tree(iter)
    self.update_options_frame()

    name = self.get_spudpath(active_tree)
    self.statusbar.set_statusbar(name)
    self.current_spudpath = name
    self.current_xpath = self.get_xpath(active_tree)

    self.clear_plugin_buttons()

    for plugin in plugins.plugins:
      if plugin.matches(name):
        self.add_plugin_button(plugin)

    return

  def get_spudpath(self, active_tree):
    # get the name to paint on the statusbar
    name_tree = active_tree
    name = ""
    while name_tree is not None:
      if "name" in name_tree.attrs and name_tree.attrs["name"][1] is not None:
        used_name = name_tree.name + '::%s' % name_tree.attrs["name"][1]
      elif name_tree.parent is not None and name_tree.parent.count_children_by_schemaname(name_tree.schemaname) > 1:
        siblings = [x for x in name_tree.parent.children if x.schemaname == name_tree.schemaname]
        i = 0
        for sibling in siblings:
          if sibling is name_tree:
            break
          else:
            i = i + 1
        used_name = name_tree.name + "[%s]" % i
      else:
        used_name = name_tree.name

      name = "/" + used_name + name
      name_tree = name_tree.parent

    # and split off the root name:
    name = '/' + '/'.join(name.split('/')[2:])
    return name

  def get_xpath(self, active_tree):
    # get the name to paint on the statusbar
    name_tree = active_tree
    name = ""
    while name_tree is not None:
      if "name" in name_tree.attrs and name_tree.attrs["name"][1] is not None:
        used_name = name_tree.name + '[@name="%s"]' % name_tree.attrs["name"][1]
      elif name_tree.parent is not None and name_tree.parent.count_children_by_schemaname(name_tree.schemaname) > 1:
        siblings = [x for x in name_tree.parent.children if x.schemaname == name_tree.schemaname]
        i = 0
        for sibling in siblings:
          if sibling is name_tree:
            break
          else:
            i = i + 1
        used_name = name_tree.name + "[%s]" % i
      else:
        used_name = name_tree.name

      name = "/" + used_name + name
      name_tree = name_tree.parent

    return name

  def clear_plugin_buttons(self):
    for button in self.plugin_buttons:
      self.plugin_buttonbox.remove(button)
    
    self.plugin_buttons = []

  def add_plugin_button(self, plugin):
    button = gtk.Button(label=plugin.name)
    button.connect('clicked', self.plugin_handler, plugin)
    button.show()

    self.plugin_buttons.append(button)
    self.plugin_buttonbox.add(button)

  def plugin_handler(self, widget, plugin):
    f = StringIO.StringIO()
    self.tree.write(f)
    xml = f.getvalue()
    xml = plugin.execute(xml, self.current_xpath)
    if xml:
      ios = StringIO.StringIO(xml)

      try:
        tree_read = self.s.read(filename)

        if tree_read is None:
          self.statusbar.set_statusbar("Unable to read plugin result")
          return

        self.display_validation_errors(self.s.read_errors())
        self.tree = tree_read
      except:
        dialogs.error_tb(self.main_window, "Unable to read plugin result")
        return

      path = self.treestore.get_path(self.selected_iter)

      self.display_validation_errors(self.s.read_errors())

      self.set_saved(False)
 
      self.treeview.freeze_child_notify()
      self.treeview.set_model(None)
      self.signals = {}
      self.set_treestore(None, [self.tree], True)
      self.treeview.set_model(self.treestore)
      self.treeview.thaw_child_notify()

      self.set_geometry_dim_tree()
  
      self.treeview.get_selection().select_path(path)

      self.scherror.destroy_error_list()


  def get_selected_row(self, selection=None):
    """
    Get the iter to the selected row.
    """

    if (selection == None):
        selection = self.treeview.get_selection()

    (model, paths) = selection.get_selected_rows()
    if paths:
      return paths[0]
    else:
      return None

  def on_activate_row(self, treeview, path, view_column):
    """
    Called when you double click or press Enter on a row.
    """

    iter = self.treestore.get_iter(path)
    
    self.expand_tree(iter)
    self.on_select_row()

    if path is None: 
      return
    
    if treeview.row_expanded(path):
      treeview.collapse_row(path)
    else:
      treeview.expand_row(path, False)

    return

  def cellcombo_changed(self, combo, tree_path, combo_iter):
    """
    This is called when a cellcombo on the left-hand treeview is edited,
    i.e. the user chooses between more than one possible choice.
    """

    tree_iter = self.treestore.get_iter(tree_path)
    choice = self.treestore.get_value(tree_iter, 0)

    # get the ref to the new active choice
    liststore = self.treestore.get_value(tree_iter, 2)
    ref = liststore.get_value(combo_iter, 1)

    # record the choice in the datatree
    choice.set_active_choice_by_ref(ref)
    new_active_tree = choice.get_current_tree()

    name = self.get_spudpath(new_active_tree)
    self.statusbar.set_statusbar(name)
    self.treestore.set(tree_iter, 1, new_active_tree)
    self.current_spudpath = name
    xpath = self.get_xpath(new_active_tree)
    self.current_xpath = xpath

    self.clear_plugin_buttons()

    for plugin in plugins.plugins:
      if plugin.matches(xpath):
        self.add_plugin_button(plugin)

    self.remove_children(tree_iter)
    self.expand_treestore(tree_iter)
    self.treeview.expand_row(tree_path, False)

    self.set_saved(False)
    self.selected_node = self.get_painted_tree(tree_iter)
    self.update_options_frame()

    return

  def get_treeview_iter(self, selection):
    """
    Get a treeview iterator object, given a selection.
    """

    path = self.get_selected_row(selection)
    if path is None:
      return self.get_treestore_iter_from_xmlpath(self.current_xpath)

    return self.treestore.get_iter(path)

  def on_set_data(self, node, data, path):
    self.set_saved(False)
    self.treeview.queue_draw()

  def on_set_attr(self, node, attr, value, path):
    self.set_saved(False)
    self.treeview.queue_draw()

  def get_painted_tree(self, iter_or_tree, lock_geometry_dim = True):
    """
    Check if the given tree, or the active tree at the given iter in the treestore,
    have any children of the form *_value. If so, we need to make the node painted
    by the options tree a mix of the two: the documentation and attributes come from
    the parent, and the data from the child.

    Also check if it is the geometry node, validity of any tuple data, and, if an
    iter is supplied, check that the node is active.
    """

    if isinstance(iter_or_tree, tree.Tree):
      active_tree = iter_or_tree
    else:
      active_tree = self.treestore.get_value(iter_or_tree, 1)

    painted_tree = active_tree.get_mixed_data()

    if not isinstance(iter_or_tree, tree.Tree) and not self.iter_is_active(self.treestore, iter_or_tree):
      painted_tree = tree.Tree(painted_tree.name, painted_tree.schemaname, painted_tree.attrs, doc = painted_tree.doc)
      painted_tree.active = False
    elif lock_geometry_dim and self.geometry_dim_tree is not None and self.geometry_dim_tree.data is not None:
      if active_tree is self.geometry_dim_tree:
        data_tree = tree.Tree(painted_tree.name, painted_tree.schemaname, datatype = "fixed")
        data_tree.data = painted_tree.data
        painted_tree = MixedTree(painted_tree, data_tree)
      elif isinstance(self.geometry_dim_tree, mixedtree.MixedTree) and active_tree is self.geometry_dim_tree.parent:
        data_tree = tree.Tree(painted_tree.child.name, painted_tree.child.schemaname, datatype = "fixed")
        data_tree.data = painted_tree.data
        painted_tree = mixedtree.MixedTree(painted_tree, data_tree)

    return painted_tree

  def get_treestore_iter_from_xmlpath(self, xmlpath):
    """
    Convert the given XML path to an iter into the treestore. For children of a
    single parent with the same names, only the first child is considered.
    """

    names = xmlpath.split("/")

    iter = self.treestore.get_iter_first()
    if iter is None:
      return None
    for name in names[1:-1]:
      while str(self.treestore.get_value(iter, 0)) != name:
        iter = self.treestore.iter_next(iter)
        if iter is None:
          return None
      iter = self.treestore.iter_children(iter)
      if iter is None:
        return None

    return iter

  def get_treestore_path_from_node(self, node):
    """
    Look for the path for the given node.
    """
    
    def search(iter, node, indent=""):
      while iter:
        if self.treestore.get_value(iter, 0) is node:
          return iter
        else:
          child = search(self.treestore.iter_children(iter), node, indent + "  ")
          if child: return child

          iter = self.treestore.iter_next(iter)      
      return iter

    iter = search(self.treestore.get_iter_first(), node)
    return self.treestore.get_path(iter) if iter else None
 
  def set_geometry_dim_tree(self):
    """
    Find the iter into the treestore corresponding to the geometry dimension, and
    perform checks to test that the geometry dimension node is valid.
    """
    
    self.geometry_dim_tree = self.data.geometry_dim_tree = None
    # The tree must exist
    if self.tree is None:
      return

    # A geometry dimension element must exist
    iter = self.get_treestore_iter_from_xmlpath("/" + self.tree.name + self.data_paths["dim"])
    if iter is None:
      return

    painted_tree = self.get_painted_tree(iter, False)
    if isinstance(painted_tree, mixedtree.MixedTree):
       # If the geometry dimension element has a hidden data element, it must
       # have datatype tuple or fixed
       if not isinstance(painted_tree.datatype, tuple) and painted_tree.datatype != "fixed":
         self.geometry_dim_tree = None
         return
    elif painted_tree.datatype != "fixed":
      # Otherwise, only fixed datatype is permitted
      return

    # All parents of the geometry dimension element must have cardinality ""
    # (i.e. not ?, * or +).
    parent = painted_tree.parent
    while parent is not None:
      if parent.cardinality != "":
        self.geometry_dim_tree = None
        return
      parent = parent.parent

    # All possible geometry dimensions must be positive integers
    if isinstance(painted_tree.datatype, tuple):
      possible_dims = painted_tree.datatype
    elif painted_tree.datatype == "fixed":
      possible_dims = [painted_tree.data]
    else:
      return
    for opt in possible_dims:
      try:
        test = int(opt)
        assert test > 0
      except:
        return
      
    # A valid geometry dimension element has been located
    self.geometry_dim_tree = self.data.geometry_dim_tree = painted_tree
    
    return

  def iter_is_active(self, treemodel, iter):
    """
    Test whether the node at the given iter in the LHS treestore is active.
    """

    while iter is not None:
      choice_or_tree = treemodel.get_value(iter, 0)
      active_tree = treemodel.get_value(iter, 1)
      if not choice_or_tree.active or not active_tree.active:
        return False
      iter = treemodel.iter_parent(iter)

    return True

  def choice_or_tree_matches(self, text, choice_or_tree, recurse, search_active_subtrees = False):
    """
    See if the supplied node matches a given piece of text. If recurse is True,
    the node is deemed to match if any of its children match or, if the node is a
    choice and search_active_subtrees is True, any of the available trees in the
    choice match.
    """

    if choice_or_tree.is_hidden():
      return False
    elif isinstance(choice_or_tree, choice.Choice):
      if self.choice_or_tree_matches(text, choice_or_tree.get_current_tree(), False):
        return True
      elif recurse and self.find.search_gui.get_widget("searchInactiveChoiceSubtreesCheckButton").get_active():
        for opt in choice_or_tree.choices():
          if not search_active_subtrees and opt is choice_or_tree.get_current_tree():
            continue
          if opt.children == []:
            self.expand_choice_or_tree(opt)
          if self.choice_or_tree_matches(text, opt, recurse, True):
            return True
    else:
      if self.get_painted_tree(choice_or_tree).matches(text, self.find.search_gui.get_widget("caseSensitiveCheckButton").get_active()):
        return True
      else:
        if self.find.search_gui.get_widget("caseSensitiveCheckButton").get_active():
          text_re = re.compile(text)
        else:
          text_re = re.compile(text, re.IGNORECASE)
        comment = choice_or_tree.get_comment()
        if comment is not None and comment.data is not None and text_re.search(comment.data) is not None:
          return True
        elif recurse:
          for opt in choice_or_tree.children:
            if self.choice_or_tree_matches(text, opt, recurse, True):
              return True

      return False

  def search_treestore(self, text, iter = None):
    """
    Recursively search the tree for a node that matches a given piece of text.
    MixedTree.matches and choice_or_tree_matches decide what is a match (using
    tree.Matches).

    This uses lazy evaluation to only search as far as necessary; I love
    Python generators. If you don't know what a Python generator is and need to
    understand this, see PEP 255.
    """

    if iter is None:
      iter = self.treestore.get_iter_first()
    if iter is None:
      yield None
    choice_or_tree = self.treestore.get_value(iter, 0)

    if self.choice_or_tree_matches(text, choice_or_tree, isinstance(choice_or_tree, choice.Choice)):
      yield iter

    child_iter = self.treestore.iter_children(iter)
    while child_iter is not None:
      for iter in self.search_treestore(text, child_iter):
        yield iter
      child_iter = self.treestore.iter_next(child_iter)

    return

  ### RHS ###

  def add_custom_widgets(self):
    """
    Adds custom python widgets that aren't easily handeled by glade.
    """
    
    optionsFrame = self.gui.get_widget("optionsFrame")

    vpane1 = gtk.VPaned()
    vpane2 = gtk.VPaned()
    vbox = gtk.VBox()
    
    vpane1.pack2(vpane2, True, False)
    vpane2.pack1(vbox, True, False)
    optionsFrame.add(vpane1)

    self.description = descriptionwidget.DescriptionWidget()
    vpane1.pack1(self.description, True, False)

    self.attributes = attributewidget.AttributeWidget()
    vbox.pack_start(self.attributes, True, True)

    databuttons = databuttonswidget.DataButtonsWidget()
    vbox.pack_end(databuttons, False)
    
    self.data = datawidget.DataWidget()
    self.data.set_buttons(databuttons)
    vbox.pack_end(self.data, True, True)

    self.comment = commentwidget.CommentWidget()
    vpane2.pack2(self.comment, True, False)

    optionsFrame.show_all()
    return 
 
  def update_options_frame(self):
    """
    Update the RHS.
    """
    
    self.description.update(self.selected_node)

    self.attributes.update(self.selected_node)

    self.data.update(self.selected_node)

    self.comment.update(self.selected_node)

    self.gui.get_widget("optionsFrame").queue_resize()
    
    return

class DiamondFindDialog:
  def __init__(self, parent, gladefile):
    self.parent = parent
    self.gladefile = gladefile
    self.search_dialog = None

    return

  def on_find(self, widget=None):
    """
    Open up the find dialog. It has to be created each time from the glade file.
    """

    if not self.search_dialog is None:
      return

    signals =      {"on_find_dialog_close": self.on_find_close_button,
                    "on_close_clicked": self.on_find_close_button,
                    "on_find_clicked": self.on_find_find_button}

    self.search_gui = gtk.glade.XML(self.gladefile, root="find_dialog")
    self.search_dialog = self.search_gui.get_widget("find_dialog")
    self.search_gui.signal_autoconnect(signals)
    search_entry = self.search_gui.get_widget("search_entry")
    search_entry.connect("activate", self.on_find_find_button)

    # reset the search parameters
    self.search_generator = None
    self.search_text = ""
    self.search_count = 0
    self.search_dialog.show()
    self.parent.statusbar.set_statusbar("")

    return

  def on_find_find_button(self, button):
    """
    Search. Each time "Find" is clicked, we compare the stored search text to the
    text in the entry box. If it's the same, we find next; if it's different, we
    start a new search. self.search_treestore does the heavy lifting.
    """

    search_entry = self.search_gui.get_widget("search_entry")

    self.parent.statusbar.clear_statusbar()

    text = search_entry.get_text()
    if text == "":
      self.parent.statusbar.set_statusbar("No text")
      return

    # check if we've started a new search
    if text != self.search_text:
      # started a new search
      self.search_generator = None
      self.search_generator = self.parent.search_treestore(text)
      self.search_text = text
      self.search_count = 0

    try:
      # get the iter of the next tree that matches
      iter = self.search_generator.next()
      path = self.parent.treestore.get_path(iter)
      # scroll down to it, expand it, and select it
      self.parent.treeview.expand_to_path(path)
      self.parent.treeview.get_selection().select_iter(iter)
      self.parent.treeview.scroll_to_cell(path, use_align=True, col_align=0.5)
      # count how many hits we've had
      self.search_count = self.search_count + 1
    except StopIteration:
      # reset the search and cycle
      self.search_text = ""
      # if something was found, go through again
      if self.search_count > 0:
        self.on_find_find_button(button)
      else:
        self.parent.statusbar.set_statusbar("No results")

    return

  def on_find_close_button(self, button = None):
    """
    Close the search widget.
    """

    if not self.search_dialog is None:
      self.search_dialog.hide()
      self.search_dialog = None
    self.parent.statusbar.clear_statusbar()

    return

class DiamondStatusBar:
  def __init__(self, statusbar):
    self.statusbar = statusbar
    self.context_id = statusbar.get_context_id("Messages")

    return

  def set_statusbar(self, msg):
    """
    Set the status bar message.
    """

    self.statusbar.push(self.context_id, msg)

    return

  def clear_statusbar(self):
    """
    Clear the status bar.
    """

    self.statusbar.push(self.context_id, "")

    return
